@@ -36,15 +36,30 @@ G_BEGIN_DECLS
 typedef struct _FmPath FmPath;
 typedef struct _FmPathList FmPathList;
 
+/**
+ * FmPathFlags:
+ * @FM_PATH_NONE: -
+ * @FM_PATH_IS_NATIVE: This is a native path to UNIX, like /home
+ * @FM_PATH_IS_LOCAL: This path refers to a file on local filesystem
+ * @FM_PATH_IS_VIRTUAL: This path is virtual and it doesn't exist on real filesystem
+ * @FM_PATH_IS_TRASH: This path is under trash:///
+ * @FM_PATH_IS_XDG_MENU: This path is under menu:///
+ *
+ * Flags of #FmPath object.
+ *
+ * FM_PATH_IS_VIRTUAL and FM_PATH_IS_XDG_MENU are deprecated since 1.0.2
+ * and should not be used in newly written code.
+ */
 typedef enum
 {
     FM_PATH_NONE = 0,
-    FM_PATH_IS_NATIVE = 1<<0, /* This is a native path to UNIX, like /home */
-    FM_PATH_IS_LOCAL = 1<<1, /* This path refers  to a file on local filesystem */
-    FM_PATH_IS_VIRTUAL = 1<<2, /* This path is virtual and it doesn't exist on real filesystem */
-    FM_PATH_IS_TRASH = 1<<3, /* This path is under trash:/// */
-    FM_PATH_IS_XDG_MENU = 1<<4, /* This path is under menu:/// */
+    FM_PATH_IS_NATIVE = 1<<0,
+    FM_PATH_IS_LOCAL = 1<<1,
+    FM_PATH_IS_VIRTUAL = 1<<2,
+    FM_PATH_IS_TRASH = 1<<3,
+    FM_PATH_IS_XDG_MENU = 1<<4,
 
+    /*< private >*/
     /* reserved for future use */
     FM_PATH_IS_RESERVED1 = 1<<5,
     FM_PATH_IS_RESERVED2 = 1<<6,
@@ -53,20 +68,14 @@ typedef enum
 
 typedef struct _FmFileInfoList FmFileInfoList; /* fm-file-info.h includes this too */
 
-struct _FmPath
-{
-    gint n_ref;
-    FmPath* parent;
-    guchar flags; /* FmPathFlags flags : 8; */
-    char name[1];
-};
-
 void _fm_path_init(void);
 void _fm_path_finalize(void);
 
 FmPath* fm_path_new_for_path(const char* path_name);
 FmPath* fm_path_new_for_uri(const char* uri);
+#ifndef FM_DISABLE_DEPRECATED
 FmPath* fm_path_new_for_display_name(const char* path_name);
+#endif
 FmPath* fm_path_new_for_str(const char* path_str);
 FmPath* fm_path_new_for_commandline_arg(const char* arg);
 
@@ -93,9 +102,11 @@ gboolean fm_path_has_prefix(FmPath* path, FmPath* prefix);
 #define fm_path_is_native(path) ((fm_path_get_flags(path)&FM_PATH_IS_NATIVE)!=0)
 #define fm_path_is_trash(path) ((fm_path_get_flags(path)&FM_PATH_IS_TRASH)!=0)
 #define fm_path_is_trash_root(path) (path == fm_path_get_trash())
-#define fm_path_is_virtual(path) ((fm_path_get_flags(path)&FM_PATH_IS_VIRTUAL)!=0)
 #define fm_path_is_local(path) ((fm_path_get_flags(path)&FM_PATH_IS_LOCAL)!=0)
+#ifndef FM_DISABLE_DEPRECATED
+#define fm_path_is_virtual(path) ((fm_path_get_flags(path)&FM_PATH_IS_VIRTUAL)!=0)
 #define fm_path_is_xdg_menu(path) ((fm_path_get_flags(path)&FM_PATH_IS_XDG_MENU)!=0)
+#endif
 
 char* fm_path_to_str(FmPath* path);
 char* fm_path_to_uri(FmPath* path);
@@ -108,6 +119,9 @@ char* fm_path_display_basename(FmPath* path);
 guint fm_path_hash(FmPath* path);
 gboolean fm_path_equal(FmPath* p1, FmPath* p2);
 
+/* can be used for sorting */
+int fm_path_compare(FmPath* p1, FmPath* p2);
+
 /* used for completion in fm_path_entry */
 gboolean fm_path_equal_str(FmPath *path, const gchar *str, int n);
 
