@@ -20,6 +20,16 @@
  *      MA 02110-1301, USA.
  */
 
+/**
+ * SECTION:fm-utils
+ * @short_description: Common utility functions used by libfm and libfm-gtk.
+ * @title: Common Libfm utilities.
+ *
+ * @include: libfm/fm-utils.h
+ *
+ * This scope contains common data parsing utilities.
+ */
+
 #ifdef HAVE_CONFIG_H
 #include <config.h>
 #endif
@@ -44,6 +54,19 @@
 #define SI_GB   ((gdouble)1000.0 * 1000.0 * 1000.0)
 #define SI_TB   ((gdouble)1000.0 * 1000.0 * 1000.0 * 1000.0)
 
+/**
+ * fm_file_size_to_str
+ * @buf: pointer to array to make a string
+ * @buf_size: size of @buf
+ * @size: number to convert
+ * @si_prefix: %TRUE to convert in SI units, %FALSE to convert in IEC units
+ *
+ * Converts @size into text representation of form "21.4 kiB" for example.
+ *
+ * Returns: @buf.
+ *
+ * Since: 0.1.0
+ */
 char* fm_file_size_to_str( char* buf, size_t buf_size, goffset size, gboolean si_prefix )
 {
     const char * unit;
@@ -115,6 +138,19 @@ char* fm_file_size_to_str( char* buf, size_t buf_size, goffset size, gboolean si
     return buf;
 }
 
+/**
+ * fm_key_file_get_int
+ * @kf: a key file
+ * @grp: group to lookup key
+ * @key: a key to lookup
+ * @val: location to store value
+ *
+ * Lookups @key in @kf and stores found value in @val if the @key was found.
+ *
+ * Returns: %TRUE if @key was found.
+ *
+ * Since: 0.1.0
+ */
 gboolean fm_key_file_get_int(GKeyFile* kf, const char* grp, const char* key, int* val)
 {
     char* str = g_key_file_get_value(kf, grp, key, NULL);
@@ -126,6 +162,19 @@ gboolean fm_key_file_get_int(GKeyFile* kf, const char* grp, const char* key, int
     return str != NULL;
 }
 
+/**
+ * fm_key_file_get_bool
+ * @kf: a key file
+ * @grp: group to lookup key
+ * @key: a key to lookup
+ * @val: location to store value
+ *
+ * Lookups @key in @kf and stores found value in @val if the @key was found.
+ *
+ * Returns: %TRUE if @key was found.
+ *
+ * Since: 0.1.0
+ */
 gboolean fm_key_file_get_bool(GKeyFile* kf, const char* grp, const char* key, gboolean* val)
 {
     char* str = g_key_file_get_value(kf, grp, key, NULL);
@@ -137,6 +186,18 @@ gboolean fm_key_file_get_bool(GKeyFile* kf, const char* grp, const char* key, gb
     return str != NULL;
 }
 
+/**
+ * fm_canonicalize_filename
+ * @filename: a filename
+ * @cwd: current work directory path
+ *
+ * Makes a canonical name with full path from @filename. Returned string
+ * should be freed by caller after usage.
+ *
+ * Returns: (transfer full): a canonical name.
+ *
+ * Since: 0.1.0
+ */
 char* fm_canonicalize_filename(const char* filename, const char* cwd)
 {
     char* _cwd = NULL;
@@ -225,6 +286,22 @@ char* fm_canonicalize_filename(const char* filename, const char* cwd)
     return ret;
 }
 
+/**
+ * fm_strdup_replace
+ * @str: a string
+ * @old_str: substring to replace
+ * @new_str: string to replace @old_str
+ *
+ * Replaces every occurence of @old_str in @str with @new_str and returns
+ * resulted string. Returned string should be freed with g_free() after
+ * usage.
+ *
+ * Before 1.0.0 this API had name fm_str_replace.
+ *
+ * Returns: (transfer full): resulted string.
+ *
+ * Since: 0.1.16
+ */
 char* fm_strdup_replace(char* str, char* old_str, char* new_str)
 {
     int len = strlen(str);
@@ -245,13 +322,15 @@ char* fm_strdup_replace(char* str, char* old_str, char* new_str)
  * fm_app_command_parse
  * @cmd:        line to parse
  * @opts:       plain list of possible options
- * @ret:        pointer for resulting string, string should be freed by caller
+ * @ret:        (out) (transfer full): pointer for resulting string, string should be freed by caller
  * @user_data:  caller data to pass to callback
  *
  * This function parses line that contains some %&lt;char&gt; commands and does
  * substitutions on them using callbacks provided by caller.
  *
- * Return value: number of valid options found in @cmd
+ * Returns: number of valid options found in @cmd.
+ *
+ * Since: 1.0.0
  */
 int fm_app_command_parse(const char* cmd, const FmAppCommandParseOption* opts,
                          char** ret, gpointer user_data)
@@ -301,3 +380,112 @@ int fm_app_command_parse(const char* cmd, const FmAppCommandParseOption* opts,
     *ret = g_string_free(buf, FALSE);
     return hits;
 }
+
+
+/* ---- run menu cache in main context ---- */
+/* Lock for main context run. */
+#if GLIB_CHECK_VERSION(2, 32, 0)
+static GMutex main_loop_run_mutex;
+static GCond main_loop_run_cond;
+#else
+static GMutex *main_loop_run_mutex = NULL;
+static GCond *main_loop_run_cond = NULL;
+#endif
+
+typedef struct
+{
+    gboolean done;
+    GSourceFunc func;
+    gpointer data;
+    gboolean result;
+} _main_context_data;
+
+static gboolean _fm_run_in_default_main_context_real(gpointer user_data)
+{
+    _main_context_data *data = user_data;
+    data->result = data->func(data->data);
+#if GLIB_CHECK_VERSION(2, 32, 0)
+    g_mutex_lock(&main_loop_run_mutex);
+    data->done = TRUE;
+    g_cond_broadcast(&main_loop_run_cond);
+    g_mutex_unlock(&main_loop_run_mutex);
+#else
+    g_mutex_lock(main_loop_run_mutex);
+    data->done = TRUE;
+    g_cond_broadcast(main_loop_run_cond);
+    g_mutex_unlock(main_loop_run_mutex);
+#endif
+    return FALSE;
+}
+
+/**
+ * fm_run_in_default_main_context
+ * @func: function to run
+ * @data: data supplied for @func
+ *
+ * Runs @func once in global main loop with supplied @data.
+ *
+ * Returns: output of @func.
+ *
+ * Since: 1.0.2
+ */
+gboolean fm_run_in_default_main_context(GSourceFunc func, gpointer data)
+{
+    _main_context_data md;
+
+#if GLIB_CHECK_VERSION(2, 32, 0)
+    md.done = FALSE;
+    md.func = func;
+    md.data = data;
+    g_main_context_invoke(NULL, _fm_run_in_default_main_context_real, &md);
+    g_mutex_lock(&main_loop_run_mutex);
+    while(!md.done)
+        g_cond_wait(&main_loop_run_cond, &main_loop_run_mutex);
+    g_mutex_unlock(&main_loop_run_mutex);
+#else
+    /* if we already in main loop then just run it */
+    if(g_main_context_is_owner(g_main_context_default()))
+        md.result = func(data);
+    /* if we can acquire context then do it */
+    else if(g_main_context_acquire(g_main_context_default()))
+    {
+        md.result = func(data);
+        g_main_context_release(g_main_context_default());
+    }
+    /* else add idle source and wait for return */
+    else
+    {
+        if(!main_loop_run_mutex)
+            main_loop_run_mutex = g_mutex_new();
+        if(!main_loop_run_cond)
+            main_loop_run_cond = g_cond_new();
+        md.done = FALSE;
+        md.func = func;
+        md.data = data;
+        g_idle_add(_fm_run_in_default_main_context_real, &md);
+        g_mutex_lock(main_loop_run_mutex);
+        while(!md.done)
+            g_cond_wait(main_loop_run_cond, main_loop_run_mutex);
+        g_mutex_unlock(main_loop_run_mutex);
+    }
+#endif
+    return md.result;
+}
+
+/**
+ * fm_get_home_dir
+ *
+ * Retrieves valid path to home dir of user.
+ *
+ * Returns: path string.
+ *
+ * Since: 1.0.2
+ */
+const char *fm_get_home_dir(void)
+{
+    /* From GLib docs for g_get_home_dir */
+    const char *homedir = g_getenv("HOME");
+    if(!homedir)
+        homedir = g_get_home_dir();
+    return homedir;
+}
