@@ -2,6 +2,7 @@
  *      fm-cell-renderer-text.c
  *
  *      Copyright 2009 PCMan <pcman.tw@gmail.com>
+ *      Copyright 2012 Andriy Grytsenko (LStranger) <andrej@rep.kiev.ua>
  *
  *      This program is free software; you can redistribute it and/or modify
  *      it under the terms of the GNU General Public License as published by
@@ -31,6 +32,17 @@
 
 #include "fm-cell-renderer-text.h"
 
+static void fm_cell_renderer_text_get_property(GObject *object, guint param_id,
+                                               GValue *value, GParamSpec *psec);
+static void fm_cell_renderer_text_set_property(GObject *object, guint param_id,
+                                               const GValue *value, GParamSpec *psec);
+
+enum
+{
+    PROP_HEIGHT = 1,
+    N_PROPS
+};
+
 G_DEFINE_TYPE(FmCellRendererText, fm_cell_renderer_text, GTK_TYPE_CELL_RENDERER_TEXT);
 
 static void fm_cell_renderer_text_get_size(GtkCellRenderer            *cell,
@@ -61,17 +73,63 @@ static void fm_cell_renderer_text_render(GtkCellRenderer *cell,
                                          GtkCellRendererState  flags);
 #endif
 
+#if GTK_CHECK_VERSION(3, 0, 0)
+static void fm_cell_renderer_text_get_preferred_width(GtkCellRenderer *cell,
+                                                      GtkWidget *widget,
+                                                      gint *minimum_size,
+                                                      gint *natural_size);
+
+static void fm_cell_renderer_text_get_preferred_height(GtkCellRenderer *cell,
+                                                       GtkWidget *widget,
+                                                       gint *minimum_size,
+                                                       gint *natural_size);
+
+static void fm_cell_renderer_text_get_preferred_height_for_width(GtkCellRenderer *cell,
+                                                                 GtkWidget *widget,
+                                                                 gint width,
+                                                                 gint *minimum_height,
+                                                                 gint *natural_height);
+#endif
+
 static void fm_cell_renderer_text_class_init(FmCellRendererTextClass *klass)
 {
+    GObjectClass *g_object_class = G_OBJECT_CLASS(klass);
     GtkCellRendererClass* render_class = GTK_CELL_RENDERER_CLASS(klass);
+
+    g_object_class->get_property = fm_cell_renderer_text_get_property;
+    g_object_class->set_property = fm_cell_renderer_text_set_property;
+
     render_class->render = fm_cell_renderer_text_render;
     render_class->get_size = fm_cell_renderer_text_get_size;
+#if GTK_CHECK_VERSION(3, 0, 0)
+    render_class->get_preferred_width = fm_cell_renderer_text_get_preferred_width;
+    render_class->get_preferred_height = fm_cell_renderer_text_get_preferred_height;
+    render_class->get_preferred_height_for_width = fm_cell_renderer_text_get_preferred_height_for_width;
+#endif
+
+    /**
+     * FmCellRendererText:max-height:
+     *
+     * The #FmCellRendererText:max-height property is the maximum height
+     * of text that will be rendered. If text doesn't fit in that height
+     * then text will be ellipsized. The value of -1 means unlimited
+     * height.
+     *
+     * Since: 1.0.2
+     */
+    g_object_class_install_property(g_object_class,
+                                    PROP_HEIGHT,
+                                    g_param_spec_int("max-height",
+                                                     "Maximum_height",
+                                                     "Maximum height",
+                                                     -1, 2048, -1,
+                                                     G_PARAM_READWRITE));
 }
 
 
 static void fm_cell_renderer_text_init(FmCellRendererText *self)
 {
-    
+    self->height = -1;
 }
 
 /**
@@ -88,6 +146,36 @@ GtkCellRenderer *fm_cell_renderer_text_new(void)
     return (GtkCellRenderer*)g_object_new(FM_CELL_RENDERER_TEXT_TYPE, NULL);
 }
 
+static void fm_cell_renderer_text_get_property(GObject *object, guint param_id,
+                                               GValue *value, GParamSpec *psec)
+{
+    FmCellRendererText *self = FM_CELL_RENDERER_TEXT(object);
+    switch(param_id)
+    {
+    case PROP_HEIGHT:
+        g_value_set_int(value, self->height);
+        break;
+    default:
+        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, param_id, psec);
+        break;
+    }
+}
+
+static void fm_cell_renderer_text_set_property(GObject *object, guint param_id,
+                                               const GValue *value, GParamSpec *psec)
+{
+    FmCellRendererText *self = FM_CELL_RENDERER_TEXT(object);
+    switch(param_id)
+    {
+    case PROP_HEIGHT:
+        self->height = g_value_get_int(value);
+        break;
+    default:
+        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, param_id, psec);
+        break;
+    }
+}
+
 #if GTK_CHECK_VERSION(3, 0, 0)
 static void fm_cell_renderer_text_render(GtkCellRenderer *cell,
                                          cairo_t *cr,
@@ -105,6 +193,7 @@ static void fm_cell_renderer_text_render(GtkCellRenderer *cell,
                                          GtkCellRendererState flags)
 #endif
 {
+    FmCellRendererText *self = FM_CELL_RENDERER_TEXT(cell);
 #if GTK_CHECK_VERSION(3, 0, 0)
     GtkStyleContext* style;
     GtkStateFlags state;
@@ -121,7 +210,6 @@ static void fm_cell_renderer_text_render(GtkCellRenderer *cell,
     GdkRectangle rect;
     PangoWrapMode wrap_mode;
     gint wrap_width;
-    gint wrap_height = -1;
     PangoAlignment alignment;
     gfloat xalign, yalign;
     gint xpad, ypad;
@@ -152,12 +240,14 @@ static void fm_cell_renderer_text_render(GtkCellRenderer *cell,
     {
         pango_layout_set_width(layout, wrap_width * PANGO_SCALE);
         pango_layout_set_wrap(layout, wrap_mode);
-        /* FIXME: add custom ellipsize from object? */
-        pango_layout_set_ellipsize(layout, PANGO_ELLIPSIZE_END);
-        /* Setup max height to be not more than half of width */
-        wrap_height = wrap_width * PANGO_SCALE / 2;
-        /* FIXME: add custom height from object? */
-        pango_layout_set_height(layout, wrap_height);
+        if(self->height > 0)
+        {
+            /* FIXME: add custom ellipsize from object? */
+            pango_layout_set_ellipsize(layout, PANGO_ELLIPSIZE_END);
+            pango_layout_set_height(layout, self->height * PANGO_SCALE);
+        }
+        else
+            pango_layout_set_ellipsize(layout, PANGO_ELLIPSIZE_NONE);
     }
 
     pango_layout_set_text(layout, text, -1);
@@ -257,6 +347,12 @@ static void fm_cell_renderer_text_render(GtkCellRenderer *cell,
                         rect.width, rect.height);
 #endif
     }
+
+    if(flags & GTK_CELL_RENDERER_PRELIT) /* hovered */
+        g_object_set(G_OBJECT(widget), "tooltip-text", text, NULL);
+    else
+        g_object_set(G_OBJECT(widget), "tooltip-text", NULL, NULL);
+    g_free(text);
 }
 
 static void fm_cell_renderer_text_get_size(GtkCellRenderer            *cell,
@@ -270,21 +366,59 @@ static void fm_cell_renderer_text_get_size(GtkCellRenderer            *cell,
                                            gint                       *width,
                                            gint                       *height)
 {
-    gint wrap_width;
+    FmCellRendererText *self = FM_CELL_RENDERER_TEXT(cell);
 
-    g_object_get(G_OBJECT(cell),
-                 "wrap-width", &wrap_width,
-                 NULL);
-
-    if (wrap_width <= 0)
+    GTK_CELL_RENDERER_CLASS(fm_cell_renderer_text_parent_class)->get_size(cell, widget, rectangle, x_offset, y_offset, width, height);
+    if (self->height > 0)
     {
-        GTK_CELL_RENDERER_CLASS(fm_cell_renderer_text_parent_class)->get_size(cell, widget, rectangle, x_offset, y_offset, width, height);
-    }
-    else
-    {
-        *width = wrap_width;
-        /* FIXME: add custom height from object? */
-        *height = wrap_width / 2;
+        if(*height > self->height)
+            *height = self->height;
     }
 }
 
+#if GTK_CHECK_VERSION(3, 0, 0)
+static void fm_cell_renderer_text_get_preferred_width(GtkCellRenderer *cell,
+                                                      GtkWidget *widget,
+                                                      gint *minimum_size,
+                                                      gint *natural_size)
+{
+    gint wrap_width;
+
+    g_object_get(G_OBJECT(cell), "wrap-width", &wrap_width, NULL);
+    if(wrap_width > 0)
+    {
+        if(minimum_size)
+            *minimum_size = wrap_width;
+        if(natural_size)
+            *natural_size = wrap_width;
+    }
+    else
+        GTK_CELL_RENDERER_CLASS(fm_cell_renderer_text_parent_class)->get_preferred_width(cell, widget, minimum_size, natural_size);
+}
+
+static void fm_cell_renderer_text_get_preferred_height(GtkCellRenderer *cell,
+                                                       GtkWidget *widget,
+                                                       gint *minimum_size,
+                                                       gint *natural_size)
+{
+    FmCellRendererText *self = FM_CELL_RENDERER_TEXT(cell);
+
+    GTK_CELL_RENDERER_CLASS(fm_cell_renderer_text_parent_class)->get_preferred_height(cell, widget, minimum_size, natural_size);
+    if (self->height > 0)
+    {
+        if(natural_size && *natural_size > self->height)
+            *natural_size = self->height;
+        if(minimum_size && *minimum_size > self->height)
+            *minimum_size = self->height;
+    }
+}
+
+static void fm_cell_renderer_text_get_preferred_height_for_width(GtkCellRenderer *cell,
+                                                                 GtkWidget *widget,
+                                                                 gint width,
+                                                                 gint *minimum_height,
+                                                                 gint *natural_height)
+{
+    fm_cell_renderer_text_get_preferred_height(cell, widget, minimum_height, natural_height);
+}
+#endif
