@@ -19,7 +19,16 @@
  *      MA 02110-1301, USA.
  */
 
+/**
+ * SECTION:fm-dnd-auto-scroll
+ * @short_description: Autoscroll when dragged out of widget.
+ * @title: Drag and drop autoscrolling
+ *
+ * @include: libfm/fm-dnd-auto-scroll.h
+ *
+ */
 #include "fm-dnd-auto-scroll.h"
+#include "gtk-compat.h"
 
 #define SCROLL_EDGE_SIZE 15
 
@@ -43,7 +52,9 @@ static gboolean on_auto_scroll(gpointer user_data)
     GtkWidget* widget = as->widget;
     GtkAllocation allocation;
 
-    gdk_window_get_pointer(gtk_widget_get_window(widget), &x, &y, NULL);
+    gdk_window_get_device_position (gtk_widget_get_window(widget),
+                                    gtk_get_current_event_device(),
+                                    &x, &y, NULL);
     gtk_widget_get_allocation(widget, &allocation);
 
     /*
