@@ -41,7 +41,12 @@
  * widget. The handler should check if drop can be performed. And if
  * #FmDndDest can accept the drop then widget should inform #FmDndDest
  * object about #FmFileInfo object the mouse pointer targets at that
- * moment by calling fm_dnd_dest_set_dest_file().
+ * moment by calling fm_dnd_dest_set_dest_file(). The #FmDndDest uses a
+ * little different sequence for collecting dragged data - it queries
+ * data in time of drag motion and uses when data are dropped therefore
+ * widget should always call API fm_dnd_dest_get_default_action() from
+ * handler of the #GtkWidget::drag-motion signal for any target which
+ * #FmDndDest supports.
  * <example id="example-fmdnddest-usage">
  * <title>Sample Usage</title>
  * <programlisting>
@@ -127,16 +132,12 @@ GtkTargetEntry fm_default_dnd_dest_targets[] =
     /* TODO: add more targets to support: text types, _NETSCAPE_URL, property/bgimage ... */
 };
 
-/* GdkAtom value for drag target: XdndDirectSave0 */
-static GdkAtom xds_target_atom = 0;
-#if 0
 static GdkAtom dest_target_atom[N_FM_DND_DEST_DEFAULT_TARGETS];
-#endif
 
 static void fm_dnd_dest_dispose              (GObject *object);
 static gboolean fm_dnd_dest_files_dropped(FmDndDest* dd, int x, int y, guint action, guint info_type, FmPathList* files);
 
-static gboolean clear_src_cache(gpointer user_data);
+static void clear_src_cache(FmDndDest* dd);
 
 static void on_drag_leave(GtkWidget *widget, GdkDragContext *drag_context,
                           guint time, FmDndDest* dd);
@@ -156,6 +157,7 @@ static void fm_dnd_dest_class_init(FmDndDestClass *klass)
 {
     GObjectClass *g_object_class;
     FmDndDestClass *dnd_dest_class;
+    guint i;
 
     g_object_class = G_OBJECT_CLASS(klass);
     g_object_class->dispose = fm_dnd_dest_dispose;
@@ -190,14 +192,11 @@ static void fm_dnd_dest_class_init(FmDndDestClass *klass)
                      fm_marshal_BOOL__INT_INT_UINT_UINT_POINTER,
                      G_TYPE_BOOLEAN, 5, G_TYPE_INT, G_TYPE_INT, G_TYPE_UINT, G_TYPE_UINT, G_TYPE_POINTER);
 
-    xds_target_atom = gdk_atom_intern_static_string(fm_default_dnd_dest_targets[FM_DND_DEST_TARGET_XDS].target);
-#if 0
     for(i = 0; i < N_FM_DND_DEST_DEFAULT_TARGETS; i++)
         dest_target_atom[i] = GDK_NONE;
     for(i = 0; i < G_N_ELEMENTS(fm_default_dnd_dest_targets); i++)
         dest_target_atom[fm_default_dnd_dest_targets[i].info] =
             gdk_atom_intern_static_string(fm_default_dnd_dest_targets[i].target);
-#endif
 }
 
 
@@ -212,6 +211,11 @@ static void fm_dnd_dest_dispose(GObject *object)
 
     fm_dnd_dest_set_widget(dd, NULL);
 
+    if(dd->idle)
+    {
+        g_source_remove(dd->idle);
+        dd->idle = 0;
+    }
     clear_src_cache(dd);
 
     G_OBJECT_CLASS(fm_dnd_dest_parent_class)->dispose(object);
@@ -323,6 +327,7 @@ void fm_dnd_dest_set_widget(FmDndDest* dd, GtkWidget* w)
     }
 }
 
+/* own handler for "files-dropped" signal */
 static gboolean fm_dnd_dest_files_dropped(FmDndDest* dd, int x, int y,
                                           guint action, guint info_type,
                                           FmPathList* files)
@@ -373,17 +378,15 @@ static gboolean fm_dnd_dest_files_dropped(FmDndDest* dd, int x, int y,
         break;
     case GDK_ACTION_ASK:
         g_debug("TODO: GDK_ACTION_ASK");
-        break;
     case GDK_ACTION_PRIVATE:
     default: /* invalid combination */
-        ;
+        return FALSE;
     }
     return TRUE;
 }
 
-static gboolean clear_src_cache(gpointer user_data)
+static void clear_src_cache(FmDndDest* dd)
 {
-    FmDndDest* dd = (FmDndDest*)user_data;
     /* free cached source files */
     if(dd->src_files)
     {
@@ -399,12 +402,16 @@ static gboolean clear_src_cache(gpointer user_data)
     dd->src_fs_id = NULL;
 
     dd->info_type = 0;
-    if(dd->idle)
-    {
-        g_source_remove(dd->idle);
-        dd->idle = 0;
-    }
     dd->waiting_data = FALSE;
+}
+
+static gboolean clear_src_cache_on_idle(gpointer user_data)
+{
+    GDK_THREADS_ENTER();
+    /* check if dd is still valid */
+    if(!g_source_is_destroyed(g_main_current_source()))
+        clear_src_cache((FmDndDest*)user_data);
+    GDK_THREADS_LEAVE();
     return FALSE;
 }
 
@@ -498,10 +505,9 @@ gboolean _on_drag_data_received(FmDndDest* dd, GdkDragContext *drag_context,
     data = (const gchar*)gtk_selection_data_get_data_with_length(sel_data, &length);
     format = gtk_selection_data_get_format(sel_data);
 
-    if(info ==  FM_DND_DEST_TARGET_FM_LIST)
+    if(info == FM_DND_DEST_TARGET_FM_LIST)
     {
         if((length == sizeof(gpointer)) && (format==8))
-        /* FIXME: check if it's internal within application */
         {
             /* get the pointer */
             FmFileInfoList* file_infos = *(FmFileInfoList**)data;
@@ -551,10 +557,7 @@ gboolean _on_drag_data_received(FmDndDest* dd, GdkDragContext *drag_context,
         {
             GdkWindow *source_window;
             source_window = gdk_drag_context_get_source_window(drag_context);
-            gdk_property_change(source_window, xds_target_atom,
-#if 0
             gdk_property_change(source_window, dest_target_atom[FM_DND_DEST_TARGET_XDS],
-#endif
                                gdk_atom_intern_static_string("text/plain"), 8,
                                GDK_PROP_MODE_REPLACE, (const guchar *)"", 0);
         }
@@ -565,17 +568,20 @@ gboolean _on_drag_data_received(FmDndDest* dd, GdkDragContext *drag_context,
         gtk_drag_finish(drag_context, TRUE, FALSE, time);
         return TRUE;
     }
-    else
-        return FALSE;
 
     /* remove previously cached source files. */
     if(G_UNLIKELY(dd->src_files))
         fm_path_list_unref(dd->src_files);
+    if(files && fm_path_list_is_empty(files))
+    {
+        g_warning("drag-data-received with empty list");
+        fm_path_list_unref(files);
+        files = NULL;
+    }
     dd->src_files = files;
     dd->waiting_data = FALSE;
     dd->info_type = info;
-    /* FIXME: is it normal to return TRUE while dd->src_files is NULL? */
-    return TRUE;
+    return (files != NULL);
 }
 
 gboolean fm_dnd_dest_drag_data_received(FmDndDest* dd, GdkDragContext *drag_context,
@@ -605,21 +611,13 @@ static void on_drag_data_received(GtkWidget *w, GdkDragContext *drag_context,
 GdkAtom fm_dnd_dest_find_target(FmDndDest* dd, GdkDragContext *drag_context)
 {
     guint i;
-    for(i = 0; i < G_N_ELEMENTS(fm_default_dnd_dest_targets); ++i)
-    {
-        GdkAtom target = gdk_atom_intern_static_string(fm_default_dnd_dest_targets[i].target);
-        if(fm_drag_context_has_target(drag_context, target))
-            return target;
-    }
-#if 0
-    for(i = 0; i < N_FM_DND_DEST_DEFAULT_TARGETS; i++)
+    for(i = 1; i < N_FM_DND_DEST_DEFAULT_TARGETS; i++)
     {
         GdkAtom target = dest_target_atom[i];
         if(G_LIKELY(target != GDK_NONE)
            && fm_drag_context_has_target(drag_context, target))
             return target;
     }
-#endif
     return GDK_NONE;
 }
 
@@ -636,25 +634,13 @@ GdkAtom fm_dnd_dest_find_target(FmDndDest* dd, GdkDragContext *drag_context)
  */
 gboolean fm_dnd_dest_is_target_supported(FmDndDest* dd, GdkAtom target)
 {
-    gboolean ret = FALSE;
     guint i;
 
-    for(i = 0; i < G_N_ELEMENTS(fm_default_dnd_dest_targets); ++i)
-    {
-        if(gdk_atom_intern_static_string(fm_default_dnd_dest_targets[i].target) == target)
-        {
-            ret = TRUE;
-            break;
-        }
-    }
-    return ret;
-#if 0
     if(G_LIKELY(target != GDK_NONE))
-        for(i = 0; i < N_FM_DND_DEST_DEFAULT_TARGETS; i++)
+        for(i = 1; i < N_FM_DND_DEST_DEFAULT_TARGETS; i++)
             if(dest_target_atom[i] == target)
                 return TRUE;
     return FALSE;
-#endif
 }
 
 /**
@@ -684,24 +670,13 @@ gboolean _on_drag_drop(FmDndDest* dd, GdkDragContext *drag_context,
     gboolean ret = FALSE;
     GtkWidget* dest_widget = dd->widget;
     guint i;
-    for(i = 0; i < G_N_ELEMENTS(fm_default_dnd_dest_targets); ++i)
-    {
-        if(gdk_atom_intern_static_string(fm_default_dnd_dest_targets[i].target) == target)
-        {
-            ret = TRUE;
-            i = fm_default_dnd_dest_targets[i].info;
-            break;
-        }
-    }
-#if 0
     if(G_LIKELY(target != GDK_NONE))
-        for(i = 0; i < N_FM_DND_DEST_DEFAULT_TARGETS; i++)
+        for(i = 1; i < N_FM_DND_DEST_DEFAULT_TARGETS; i++)
             if(dest_target_atom[i] == target)
             {
                 ret = TRUE;
                 break;
             }
-#endif
     if(ret) /* we support this kind of target */
     {
         if(i == FM_DND_DEST_TARGET_XDS) /* if this is XDS */
@@ -712,7 +687,7 @@ gboolean _on_drag_drop(FmDndDest* dd, GdkDragContext *drag_context,
             GdkAtom text_atom = gdk_atom_intern_static_string("text/plain");
             /* get filename from the source window */
             source_window = gdk_drag_context_get_source_window(drag_context);
-            if(gdk_property_get(source_window, xds_target_atom, text_atom,
+            if(gdk_property_get(source_window, target, text_atom,
                                 0, 1024, FALSE, NULL, NULL,
                                 &len, &data) && data)
             {
@@ -722,7 +697,7 @@ gboolean _on_drag_drop(FmDndDest* dd, GdkDragContext *drag_context,
                     FmPath* path = fm_path_new_child(fm_file_info_get_path(dest), (gchar*)data);
                     char* uri = fm_path_to_uri(path);
                     /* setup the property */
-                    gdk_property_change(source_window, xds_target_atom,
+                    gdk_property_change(source_window, target,
                                        text_atom, 8, GDK_PROP_MODE_REPLACE, (const guchar *)uri,
                                        strlen(uri) + 1);
                     fm_path_unref(path);
@@ -733,7 +708,7 @@ gboolean _on_drag_drop(FmDndDest* dd, GdkDragContext *drag_context,
             {
                 fm_show_error(GTK_WINDOW(gtk_widget_get_toplevel(dest_widget)), NULL,
                               _("XDirectSave failed."));
-                gdk_property_change(source_window, xds_target_atom,
+                gdk_property_change(source_window, target,
                                    text_atom, 8, GDK_PROP_MODE_REPLACE, (const guchar *)"", 0);
             }
             g_free(data);
@@ -742,7 +717,7 @@ gboolean _on_drag_drop(FmDndDest* dd, GdkDragContext *drag_context,
             return TRUE;
         }
 
-        /* see if the drag files are cached */
+        /* see if the dragged files are cached by "drag-motion" handler */
         if(dd->src_files)
         {
             GdkDragAction action = gdk_drag_context_get_selected_action(drag_context);
@@ -759,7 +734,8 @@ gboolean _on_drag_drop(FmDndDest* dd, GdkDragContext *drag_context,
             else
                 ret = FALSE;
         }
-        gtk_drag_finish(drag_context, ret, FALSE, time);
+        if(ret)
+            gtk_drag_finish(drag_context, ret, FALSE, time);
     }
     return ret;
 }
@@ -798,13 +774,24 @@ GdkDragAction fm_dnd_dest_get_default_action(FmDndDest* dd,
     FmPath* dest_path;
 
     if(!dest || !(dest_path = fm_file_info_get_path(dest)))
-        return 0;
+        /* query drag sources in any case */
+        goto query_sources;
+
+    /* we may have another data already so clear the cache */
+    if(dd->idle)
+    {
+        g_source_remove(dd->idle);
+        dd->idle = 0;
+        clear_src_cache(dd);
+    }
 
     /* special support for dropping onto desktop entry */
     if(fm_file_info_is_desktop_entry(dest))
     {
         GdkModifierType mask = 0;
-        gdk_window_get_pointer(gtk_widget_get_window(dd->widget), NULL, NULL, &mask);
+        gdk_window_get_device_position (gtk_widget_get_window(dd->widget),
+                                        gtk_get_current_event_device(),
+                                        NULL, NULL, &mask);
         mask &= gtk_accelerator_get_default_mod_mask();
         if(mask) /* some key is pressed! */
             return GDK_ACTION_ASK;
@@ -812,14 +799,12 @@ GdkDragAction fm_dnd_dest_get_default_action(FmDndDest* dd,
     }
 
     /* this is XDirectSave */
-    if(target == xds_target_atom)
-#if 0
     if(target == dest_target_atom[FM_DND_DEST_TARGET_XDS])
-#endif
         return GDK_ACTION_COPY;
 
     if(!dd->src_files)  /* we didn't have any data, cache it */
     {
+query_sources:
         action = 0;
         if(!dd->waiting_data) /* we're still waiting for "drag-data-received" signal */
         {
@@ -828,8 +813,7 @@ GdkDragAction fm_dnd_dest_get_default_action(FmDndDest* dd,
             dd->waiting_data = TRUE;
         }
     }
-
-    if(dd->src_files) /* we have got drag source files */
+    else /* we have got drag source files */
     {
         if(fm_path_is_trash(dest_path))
         {
@@ -839,7 +823,7 @@ GdkDragAction fm_dnd_dest_get_default_action(FmDndDest* dd,
             else /* files inside trash are read only */
                 action = 0;
         }
-        else if(fm_path_is_virtual(dest_path))
+        else if(!fm_path_is_native(dest_path))
         {
             /* computer:/// and network:/// shouldn't received dropped files. */
             /* FIXME: some special handling can be done with menu:// */
@@ -852,8 +836,10 @@ GdkDragAction fm_dnd_dest_get_default_action(FmDndDest* dd,
              * make the operation very slow. */
             gboolean same_fs;
             GdkModifierType mask = 0;
-            gdk_window_get_pointer(gtk_widget_get_window(dd->widget), NULL, NULL, &mask);
-            mask &= (GDK_SHIFT_MASK | GDK_CONTROL_MASK);
+            gdk_window_get_device_position (gtk_widget_get_window(dd->widget),
+                                            gtk_get_current_event_device(),
+                                            NULL, NULL, &mask);
+            mask &= gtk_accelerator_get_default_mod_mask();
             /* use Shift for Move, Ctrl for Copy, Ctrl+Shift for Link */
             if(mask == (GDK_SHIFT_MASK | GDK_CONTROL_MASK))
                 action = GDK_ACTION_LINK;
@@ -861,9 +847,10 @@ GdkDragAction fm_dnd_dest_get_default_action(FmDndDest* dd,
                 action = GDK_ACTION_MOVE;
             else if(mask == GDK_CONTROL_MASK)
                 action = GDK_ACTION_COPY;
-            else
+            else if(mask != 0) /* another modifier key was pressed */
+                action = 0;
             /* FIXME: make decision based on config: Auto / Copy / Ask */
-            if(dd->src_dev || dd->src_fs_id) /* we know the device of dragged source files */
+            else if(dd->src_dev || dd->src_fs_id) /* we know the device of dragged source files */
             {
                 /* compare the device/filesystem id against that of destination file */
                 if(fm_path_is_native(dest_path))
@@ -900,12 +887,12 @@ GdkDragAction fm_dnd_dest_get_default_action(FmDndDest* dd,
 void fm_dnd_dest_drag_leave(FmDndDest* dd, GdkDragContext* drag_context, guint time)
 {
     if(dd->idle == 0)
-        dd->idle = g_idle_add_full(G_PRIORITY_LOW, clear_src_cache, dd, NULL);
+        dd->idle = g_idle_add_full(G_PRIORITY_LOW, clear_src_cache_on_idle, dd, NULL);
 }
 
 static void on_drag_leave(GtkWidget *widget, GdkDragContext *drag_context,
                           guint time, FmDndDest* dd)
 {
     if(dd->idle == 0)
-        dd->idle = g_idle_add_full(G_PRIORITY_LOW, clear_src_cache, dd, NULL);
+        dd->idle = g_idle_add_full(G_PRIORITY_LOW, clear_src_cache_on_idle, dd, NULL);
 }
