@@ -20,6 +20,15 @@
  *      MA 02110-1301, USA.
  */
 
+/**
+ * SECTION:fm-file-properties
+ * @short_description: Dialog window for changing properties of file.
+ * @title: File properties dialog
+ *
+ * @include: libfm/fm-file-properties.h
+ *
+ */
+
 #include <config.h>
 #include <glib/gi18n-lib.h>
 
@@ -36,9 +45,7 @@
 #include "fm-file-info.h"
 #include "fm-file-properties.h"
 #include "fm-deep-count-job.h"
-#include "fm-file-ops-job.h"
 #include "fm-utils.h"
-#include "fm-path.h"
 #include "fm-config.h"
 
 #include "fm-progress-dlg.h"
@@ -141,9 +148,15 @@ static gboolean on_timeout(gpointer user_data)
 {
     FmFilePropData* data = (FmFilePropData*)user_data;
     char size_str[128];
-    FmDeepCountJob* dc = data->dc_job;
-
+    FmDeepCountJob* dc;
 
+    GDK_THREADS_ENTER();
+    if(g_source_is_destroyed(g_main_current_source()))
+    {
+        GDK_THREADS_LEAVE();
+        return FALSE;
+    }
+    dc = data->dc_job;
     if(G_LIKELY(dc && !fm_job_is_cancelled(FM_JOB(dc))))
     {
         char* str;
@@ -165,6 +178,7 @@ static gboolean on_timeout(gpointer user_data)
         gtk_label_set_text(data->size_on_disk, str);
         g_free(str);
     }
+    GDK_THREADS_LEAVE();
     return TRUE;
 }
 
@@ -454,7 +468,7 @@ static void on_response(GtkDialog* dlg, int response, FmFilePropData* data)
         }
 
         /* change default application for the mime-type if needed */
-        if(data->mime_type && data->mime_type->type && data->open_with)
+        if(data->mime_type && fm_mime_type_get_type(data->mime_type) && data->open_with)
         {
             GAppInfo* app;
             gboolean default_app_changed = FALSE;
@@ -464,7 +478,7 @@ static void on_response(GtkDialog* dlg, int response, FmFilePropData* data)
             {
                 if(default_app_changed)
                 {
-                    g_app_info_set_as_default_for_type(app, data->mime_type->type, &err);
+                    g_app_info_set_as_default_for_type(app, fm_mime_type_get_type(data->mime_type), &err);
                     if(err)
                     {
                         fm_show_error(GTK_WINDOW(dlg), NULL, err->message);
@@ -623,7 +637,7 @@ static void update_permissions(FmFilePropData* data)
         return;
     }
     if(data->has_dir)
-        gtk_label_set_label(data->exec_label, _("<b>Access content:</b>"));
+        gtk_label_set_label(data->exec_label, _("<b>_Access content:</b>"));
     if(!fm_config->advanced_mode)
     {
         gtk_widget_hide(GTK_WIDGET(data->flags_label));
@@ -788,7 +802,7 @@ static void init_application_list(FmFilePropData* data)
 {
     if(data->single_type && data->mime_type)
     {
-        if(g_strcmp0(data->mime_type->type, "inode/directory"))
+        if(g_strcmp0(fm_mime_type_get_type(data->mime_type), "inode/directory"))
             fm_app_chooser_combo_box_setup_for_mime_type(data->open_with, data->mime_type);
         else /* shouldn't allow set file association for folders. */
         {
@@ -800,6 +814,17 @@ static void init_application_list(FmFilePropData* data)
     }
 }
 
+/**
+ * fm_file_properties_widget_new
+ * @files: list of files
+ * @toplevel: choose appearance of dialog
+ *
+ * Creates new dialog widget for change properties of @files.
+ *
+ * Returns: (transfer full): a new widget.
+ *
+ * Since: 0.1.0
+ */
 GtkDialog* fm_file_properties_widget_new(FmFileInfoList* files, gboolean toplevel)
 {
     GtkBuilder* builder=gtk_builder_new();
@@ -876,6 +901,17 @@ GtkDialog* fm_file_properties_widget_new(FmFileInfoList* files, gboolean topleve
     return dlg;
 }
 
+/**
+ * fm_show_file_properties
+ * @parent: a window to put dialog over it
+ * @files:list of files
+ *
+ * Creates and shows file properties dialog for @files.
+ *
+ * Returns: %TRUE.
+ *
+ * Since: 0.1.0
+ */
 gboolean fm_show_file_properties(GtkWindow* parent, FmFileInfoList* files)
 {
     GtkDialog* dlg = fm_file_properties_widget_new(files, TRUE);
