@@ -29,8 +29,52 @@
  * The #FmFolderView generic interface is used to implement folder views
  * common code including handling sorting change, and keyboard and mouse
  * buttons events.
+ *
+ * The #FmFolderView interface methods can attach context menu to widget
+ * which does folder operations and consists of items:
+ * |[
+ * CreateNew -> NewFolder
+ *              NewBlank
+ *              NewShortcut
+ *              &lt;placeholder name='ph1'/&gt;
+ * ------------------------
+ * Paste
+ * Cut
+ * Copy
+ * Del
+ * Remove
+ * FileProp
+ * ------------------------
+ * SelAll
+ * InvSel
+ * ------------------------
+ * Sort -> Asc
+ *         Desc
+ *         ----------------
+ *         ByName
+ *         ByMTime
+ *         BySize
+ *         ByType
+ *         ----------------
+ *         &lt;placeholder name='CustomSortOps'/&gt;
+ * ShowHidden
+ * Rename
+ * &lt;placeholder name='CustomFolderOps'/&gt;
+ * ------------------------
+ * &lt;placeholder name='CustomCommonOps'/&gt;
+ * ------------------------
+ * Prop
+ * ]|
+ * In created menu items 'Cut', 'Copy', 'Del', 'Remove', and 'FileProp'
+ * are hidden.
+ *
+ * Widget can modity the menu replacing placeholders and hiding or
+ * enabling existing items in it. Widget can do that in callback which
+ * is supplied for call fm_folder_view_add_popup().
+ *
+ * If click was not on widget but on some item in it then not this
+ * context menu but one with #FmFileMenu object will be opened instead.
  */
-/* FIXME: add a description of popup menu */
 
 #ifdef HAVE_CONFIG_H
 #include <config.h>
@@ -44,6 +88,8 @@
 #include "fm-file-properties.h"
 #include "fm-clipboard.h"
 #include "fm-gtk-file-launcher.h"
+#include "fm-file-menu.h"
+#include "fm-gtk-utils.h"
 
 static const char folder_popup_xml[] =
 "<popup>"
@@ -130,15 +176,15 @@ static const GtkActionEntry folder_popup_actions[]=
     {"Paste", GTK_STOCK_PASTE, NULL, "<Ctrl>V", NULL, G_CALLBACK(on_paste)},
     {"Paste2", NULL, NULL, "<Ctrl>Insert", NULL, G_CALLBACK(on_paste)},
     {"Del", GTK_STOCK_DELETE, NULL, "Delete", NULL, G_CALLBACK(on_trash)},
-    {"Del2", GTK_STOCK_DELETE, NULL, "KP_Delete", NULL, G_CALLBACK(on_trash)},
+    {"Del2", NULL, NULL, "KP_Delete", NULL, G_CALLBACK(on_trash)},
     {"Remove", GTK_STOCK_REMOVE, NULL, "<Shift>Delete", NULL, G_CALLBACK(on_rm)},
-    {"Remove2", GTK_STOCK_REMOVE, NULL, "<Shift>KP_Delete", NULL, G_CALLBACK(on_rm)},
+    {"Remove2", NULL, NULL, "<Shift>KP_Delete", NULL, G_CALLBACK(on_rm)},
     {"SelAll", GTK_STOCK_SELECT_ALL, NULL, "<Ctrl>A", NULL, G_CALLBACK(on_select_all)},
     {"InvSel", NULL, N_("_Invert Selection"), "<Ctrl>I", NULL, G_CALLBACK(on_invert_select)},
     {"Sort", NULL, N_("_Sort Files"), NULL, NULL, NULL},
-    {"Rename", NULL, N_("Rename"), NULL, NULL, G_CALLBACK(on_rename)},
-    {"Prop", GTK_STOCK_PROPERTIES, NULL, "", NULL, G_CALLBACK(on_prop)},
-    {"FileProp", GTK_STOCK_PROPERTIES, NULL, "<Alt>Return", NULL, G_CALLBACK(on_file_prop)},
+    {"Rename", NULL, N_("_Rename"), NULL, NULL, G_CALLBACK(on_rename)},
+    {"Prop", GTK_STOCK_PROPERTIES, N_("Prop_erties"), "", NULL, G_CALLBACK(on_prop)},
+    {"FileProp", GTK_STOCK_PROPERTIES, N_("Prop_erties"), "<Alt>Return", NULL, G_CALLBACK(on_file_prop)},
     {"FileProp2", NULL, NULL, "<Alt>KP_Enter", NULL, G_CALLBACK(on_file_prop)},
     {"FileProp3", NULL, NULL, "<Alt>ISO_Enter", NULL, G_CALLBACK(on_file_prop)},
     {"Menu", NULL, NULL, "Menu", NULL, G_CALLBACK(on_menu)},
@@ -148,10 +194,10 @@ static const GtkActionEntry folder_popup_actions[]=
 static GtkToggleActionEntry folder_toggle_actions[]=
 {
     {"ShowHidden", NULL, N_("Show _Hidden"), "<Ctrl>H", NULL, G_CALLBACK(on_show_hidden), FALSE},
-    {"SortPerFolder", NULL, N_("Only for this folder"), NULL,
+    {"SortPerFolder", NULL, N_("_Only for this folder"), NULL,
                 N_("Check to remember sort as folder setting rather than global one"), NULL, FALSE},
-    {"MingleDirs", NULL, N_("Mingle files and folders"), NULL, NULL, NULL, FALSE},
-    {"SortIgnoreCase", NULL, N_("Ignore name case"), NULL, NULL, NULL, TRUE}
+    {"MingleDirs", NULL, N_("Mingle _files and folders"), NULL, NULL, NULL, FALSE},
+    {"SortIgnoreCase", NULL, N_("_Ignore name case"), NULL, NULL, NULL, TRUE}
 };
 
 static const GtkRadioActionEntry folder_sort_type_actions[]=
@@ -162,10 +208,10 @@ static const GtkRadioActionEntry folder_sort_type_actions[]=
 
 static const GtkRadioActionEntry folder_sort_by_actions[]=
 {
-    {"ByName", NULL, N_("By _Name"), NULL, NULL, COL_FILE_NAME},
-    {"ByMTime", NULL, N_("By _Modification Time"), NULL, NULL, COL_FILE_MTIME},
-    {"BySize", NULL, N_("By _Size"), NULL, NULL, COL_FILE_SIZE},
-    {"ByType", NULL, N_("By File _Type"), NULL, NULL, COL_FILE_DESC}
+    {"ByName", NULL, N_("By _Name"), NULL, NULL, FM_FOLDER_MODEL_COL_NAME},
+    {"ByMTime", NULL, N_("By _Modification Time"), NULL, NULL, FM_FOLDER_MODEL_COL_MTIME},
+    {"BySize", NULL, N_("By _Size"), NULL, NULL, FM_FOLDER_MODEL_COL_SIZE},
+    {"ByType", NULL, N_("By File _Type"), NULL, NULL, FM_FOLDER_MODEL_COL_DESC}
 };
 
 G_DEFINE_INTERFACE(FmFolderView, fm_folder_view, GTK_TYPE_WIDGET);
@@ -175,6 +221,7 @@ enum
     CLICKED,
     SEL_CHANGED,
     SORT_CHANGED,
+    FILTER_CHANGED,
     //CHDIR,
     N_SIGNALS
 };
@@ -250,20 +297,43 @@ static void fm_folder_view_default_init(FmFolderViewInterface *iface)
                      g_cclosure_marshal_VOID__VOID,
                      G_TYPE_NONE, 0);
 
+    /**
+     * FmFolderView::filter-changed:
+     * @view: the widget that emitted the signal
+     *
+     * The #FmFolderView::filter-changed signal is emitted when filter
+     * of the view model got changed. It's just bouncer for the same
+     * signal of #FmFolderModel.
+     *
+     * Since: 1.0.2
+     */
+    signals[FILTER_CHANGED]=
+        g_signal_new("filter-changed",
+                     FM_TYPE_FOLDER_VIEW,
+                     G_SIGNAL_RUN_FIRST,
+                     0,
+                     /* FIXME: replace 0 after ABI change
+                     G_STRUCT_OFFSET(FmFolderViewInterface, filter_changed), */
+                     NULL, NULL,
+                     g_cclosure_marshal_VOID__VOID,
+                     G_TYPE_NONE, 0);
+
     /* FIXME: add "chdir" so main-win can connect to it and sync side-pane */
 }
 
 static void on_sort_col_changed(GtkTreeSortable* sortable, FmFolderView* fv)
 {
-    int col;
-    GtkSortType order;
-    if(gtk_tree_sortable_get_sort_column_id(sortable, &col, &order))
+    if(fm_folder_model_get_sort(FM_FOLDER_MODEL(sortable), NULL, NULL))
     {
-        FM_FOLDER_VIEW_GET_IFACE(fv)->set_sort(fv, order, col);
         g_signal_emit(fv, signals[SORT_CHANGED], 0);
     }
 }
 
+static void on_filter_changed(FmFolderModel* model, FmFolderView* fv)
+{
+    g_signal_emit(fv, signals[FILTER_CHANGED], 0);
+}
+
 /**
  * fm_folder_view_set_selection_mode
  * @fv: a widget to apply
@@ -303,30 +373,42 @@ GtkSelectionMode fm_folder_view_get_selection_mode(FmFolderView* fv)
  * @type: new mode of sorting (ascending or descending)
  * @by: criteria of sorting
  *
- * Changes sorting in the view.
+ * Changes sorting in the view. Invalid values for @type or @by are
+ * ignored (will not change sorting).
+ *
+ * Since 1.0.2 values passed to this API aren't remembered in the @fv
+ * object. If @fv has no model then this API has no effect.
+ * After the model is removed from @fv (calling fm_folder_view_set_model()
+ * with NULL) there is no possibility to recover last settings and any
+ * model added to @fv later will get defaults: FM_FOLDER_MODEL_COL_DEFAULT
+ * and FM_SORT_DEFAULT.
  *
  * Since: 0.1.0
+ *
+ * Deprecated: 1.0.2: Use fm_folder_model_set_sort() instead.
  */
-void fm_folder_view_sort(FmFolderView* fv, GtkSortType type, FmFolderModelViewCol by)
+void fm_folder_view_sort(FmFolderView* fv, GtkSortType type, FmFolderModelCol by)
 {
     FmFolderViewInterface* iface;
     FmFolderModel* model;
-    GtkSortType sort_type;
-    FmFolderModelViewCol sort_by;
+    FmSortMode mode;
 
     g_return_if_fail(FM_IS_FOLDER_VIEW(fv));
 
     iface = FM_FOLDER_VIEW_GET_IFACE(fv);
-    iface->get_sort(fv, &sort_type, &sort_by);
-    if(type == GTK_SORT_ASCENDING || type == GTK_SORT_DESCENDING)
-        sort_type = type;
-    if(FM_FOLDER_MODEL_COL_IS_VALID(by))
-        sort_by = by;
-    iface->set_sort(fv, sort_type, sort_by);
     model = iface->get_model(fv);
     if(model)
-        gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(model),
-                                             sort_by, sort_type);
+    {
+        if(type == GTK_SORT_ASCENDING || type == GTK_SORT_DESCENDING)
+        {
+            fm_folder_model_get_sort(model, NULL, &mode);
+            mode &= ~FM_SORT_ORDER_MASK;
+            mode |= (type == GTK_SORT_ASCENDING) ? FM_SORT_ASCENDING : FM_SORT_DESCENDING;
+        }
+        else
+            mode = FM_SORT_DEFAULT;
+        fm_folder_model_set_sort(model, by, mode);
+    }
     /* model will generate signal to update config if changed */
 }
 
@@ -339,16 +421,22 @@ void fm_folder_view_sort(FmFolderView* fv, GtkSortType type, FmFolderModelViewCo
  * Returns: mode of sorting (ascending or descending)
  *
  * Since: 0.1.0
+ *
+ * Deprecated: 1.0.2: use fm_folder_model_get_sort() instead.
  */
 GtkSortType fm_folder_view_get_sort_type(FmFolderView* fv)
 {
-    GtkSortType type;
-    FmFolderModelViewCol by;
+    FmFolderViewInterface* iface;
+    FmFolderModel* model;
+    FmSortMode mode;
 
     g_return_val_if_fail(FM_IS_FOLDER_VIEW(fv), GTK_SORT_ASCENDING);
 
-    FM_FOLDER_VIEW_GET_IFACE(fv)->get_sort(fv, &type, &by);
-    return type;
+    iface = FM_FOLDER_VIEW_GET_IFACE(fv);
+    model = iface->get_model(fv);
+    if(model == NULL || !fm_folder_model_get_sort(model, NULL, &mode))
+        return GTK_SORT_ASCENDING;
+    return FM_SORT_IS_ASCENDING(mode) ? GTK_SORT_ASCENDING : GTK_SORT_DESCENDING;
 }
 
 /**
@@ -360,15 +448,21 @@ GtkSortType fm_folder_view_get_sort_type(FmFolderView* fv)
  * Returns: criteria of sorting.
  *
  * Since: 0.1.0
+ *
+ * Deprecated: 1.0.2: use fm_folder_model_get_sort() instead.
  */
-FmFolderModelViewCol fm_folder_view_get_sort_by(FmFolderView* fv)
+FmFolderModelCol fm_folder_view_get_sort_by(FmFolderView* fv)
 {
-    GtkSortType type;
-    FmFolderModelViewCol by;
+    FmFolderViewInterface* iface;
+    FmFolderModel* model;
+    FmFolderModelCol by;
 
-    g_return_val_if_fail(FM_IS_FOLDER_VIEW(fv), COL_FILE_NAME);
+    g_return_val_if_fail(FM_IS_FOLDER_VIEW(fv), FM_FOLDER_MODEL_COL_DEFAULT);
 
-    FM_FOLDER_VIEW_GET_IFACE(fv)->get_sort(fv, &type, &by);
+    iface = FM_FOLDER_VIEW_GET_IFACE(fv);
+    model = iface->get_model(fv);
+    if(model == NULL || !fm_folder_model_get_sort(model, &by, NULL))
+        return FM_FOLDER_MODEL_COL_DEFAULT;
     return by;
 }
 
@@ -409,10 +503,10 @@ void fm_folder_view_set_show_hidden(FmFolderView* fv, gboolean show)
  * Retrieves setting whether hidden files in folder shown in @fv should
  * be visible or not.
  *
- * Returns: %TRUE if hidden files are visible.
- *
  * See also: fm_folder_view_set_show_hidden().
  *
+ * Returns: %TRUE if hidden files are visible.
+ *
  * Since: 0.1.0
  */
 gboolean fm_folder_view_get_show_hidden(FmFolderView* fv)
@@ -503,6 +597,7 @@ FmFolderModel* fm_folder_view_get_model(FmFolderView* fv)
 static void unset_model(FmFolderView* fv, FmFolderModel* model)
 {
     g_signal_handlers_disconnect_by_func(model, on_sort_col_changed, fv);
+    g_signal_handlers_disconnect_by_func(model, on_filter_changed, fv);
 }
 
 /**
@@ -518,21 +613,25 @@ void fm_folder_view_set_model(FmFolderView* fv, FmFolderModel* model)
 {
     FmFolderViewInterface* iface;
     FmFolderModel* old_model;
-    GtkSortType type;
-    FmFolderModelViewCol by;
+    FmFolderModelCol by = FM_FOLDER_MODEL_COL_DEFAULT;
+    FmSortMode mode = FM_SORT_ASCENDING;
 
     g_return_if_fail(FM_IS_FOLDER_VIEW(fv));
 
     iface = FM_FOLDER_VIEW_GET_IFACE(fv);
     old_model = iface->get_model(fv);
     if(old_model)
+    {
+        fm_folder_model_get_sort(old_model, &by, &mode);
         unset_model(fv, old_model);
+    }
+    /* FIXME: which setting to apply if this is first model? */
     iface->set_model(fv, model);
     if(model)
     {
-        iface->get_sort(fv, &type, &by);
-        gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(model), by, type);
+        fm_folder_model_set_sort(model, by, mode);
         g_signal_connect(model, "sort-column-changed", G_CALLBACK(on_sort_col_changed), fv);
+        g_signal_connect(model, "filter-changed", G_CALLBACK(on_filter_changed), fv);
     }
 }
 
@@ -558,13 +657,13 @@ gint fm_folder_view_get_n_selected_files(FmFolderView* fv)
  * @fv: a FmFolderView object
  *
  * Retrieves a list of
- * the currently selected files. The list should be freed after usage with
- * fm_file_info_list_unref(). If there are no files selected then return
- * value is %NULL.
+ * the currently selected files. The list should be freed after usage
+ * with fm_file_info_list_unref&lpar;). If there are no files selected then
+ * returns %NULL.
  *
  * Before 1.0.0 this API had name fm_folder_view_get_selected_files.
  *
- * Return value: (transfer full): list of selected file infos.
+ * Returns: (transfer full) (element-type FmFileInfo): list of selected file infos.
  *
  * Since: 0.1.0
  */
@@ -580,13 +679,13 @@ FmFileInfoList* fm_folder_view_dup_selected_files(FmFolderView* fv)
  * @fv: a FmFolderView object
  *
  * Retrieves a list of
- * the currently selected files. The list should be freed after usage with
- * fm_path_list_unref(). If there are no files selected then return value
- * is %NULL.
+ * the currently selected files. The list should be freed after usage
+ * with fm_path_list_unref&lpar;). If there are no files selected then returns
+ * %NULL.
  *
  * Before 1.0.0 this API had name fm_folder_view_get_selected_file_paths.
  *
- * Return value: (transfer full): list of selected file paths.
+ * Returns: (transfer full) (element-type FmPath): list of selected file paths.
  *
  * Since: 0.1.0
  */
@@ -1002,24 +1101,35 @@ static void on_menu(GtkAction* act, FmFolderView* fv)
     GtkUIManager *ui = g_object_get_qdata(G_OBJECT(fv), ui_quark);
     GtkMenu *popup = g_object_get_qdata(G_OBJECT(fv), popup_quark);
     FmFolderViewInterface *iface = FM_FOLDER_VIEW_GET_IFACE(fv);
+    FmFolderModel *model;
     gboolean show_hidden;
-    GtkSortType type;
-    FmFolderModelViewCol by;
+    FmSortMode mode;
+    GtkSortType type = GTK_SORT_ASCENDING;
+    FmFolderModelCol by;
 
+    /* FIXME: realize popup window and put it in the fv (honoring monitor) */
     /* don't show context menu outside of the folder view */
     if(act != NULL && !pointer_is_over_widget(GTK_WIDGET(fv)))
         return;
     /* FIXME: if act != NULL and there is selection in fv then open file menu instead */
     /* update actions */
-    iface->get_sort(fv, &type, &by);
+    model = iface->get_model(fv);
+    if(fm_folder_model_get_sort(model, &by, &mode))
+        type = FM_SORT_IS_ASCENDING(mode) ? GTK_SORT_ASCENDING : GTK_SORT_DESCENDING;
     act = gtk_ui_manager_get_action(ui, "/popup/Sort/Asc");
+    /* g_debug("set /popup/Sort/Asc: %u", type); */
     gtk_radio_action_set_current_value(GTK_RADIO_ACTION(act), type);
     act = gtk_ui_manager_get_action(ui, "/popup/Sort/ByName");
+    if(by == FM_FOLDER_MODEL_COL_DEFAULT)
+        by = FM_FOLDER_MODEL_COL_NAME;
+    /* g_debug("set /popup/Sort/ByName: %u", by); */
     gtk_radio_action_set_current_value(GTK_RADIO_ACTION(act), by);
     show_hidden = iface->get_show_hidden(fv);
     act = gtk_ui_manager_get_action(ui, "/popup/ShowHidden");
     gtk_toggle_action_set_active(GTK_TOGGLE_ACTION(act), show_hidden);
-    /* FIXME: shadow 'Paste' if clipboard is empty and unshadow if not */
+    /* shadow 'Paste' if clipboard is empty and unshadow if not */
+    act = gtk_ui_manager_get_action(ui, "/popup/Paste");
+    gtk_action_set_sensitive(act, fm_clipboard_have_files(GTK_WIDGET(fv)));
     /* open popup */
     gtk_menu_popup(popup, NULL, NULL, NULL, NULL, 3, gtk_get_current_event_time());
 }
@@ -1030,16 +1140,30 @@ static void on_show_hidden(GtkToggleAction* act, FmFolderView* fv)
     fm_folder_view_set_show_hidden(fv, active);
 }
 
-static void on_change_type(GtkRadioAction* act, GtkRadioAction* cur, FmFolderView* fv)
-{
-    guint val = gtk_radio_action_get_current_value(cur);
-    fm_folder_view_sort(fv, val, -1);
-}
-
 static void on_change_by(GtkRadioAction* act, GtkRadioAction* cur, FmFolderView* fv)
 {
     guint val = gtk_radio_action_get_current_value(cur);
-    fm_folder_view_sort(fv, -1, val);
+    FmFolderModel *model = fm_folder_view_get_model(fv);
+
+    /* g_debug("on_change_by"); */
+    if(model)
+        fm_folder_model_set_sort(model, val, FM_SORT_DEFAULT);
+}
+
+static void on_change_type(GtkRadioAction* act, GtkRadioAction* cur, FmFolderView* fv)
+{
+    guint val = gtk_radio_action_get_current_value(cur);
+    FmFolderModel *model = fm_folder_view_get_model(fv);
+    FmSortMode mode;
+
+    /* g_debug("on_change_type"); */
+    if(model)
+    {
+        fm_folder_model_get_sort(model, NULL, &mode);
+        mode &= ~FM_SORT_ORDER_MASK;
+        mode |= (val == GTK_SORT_ASCENDING) ? FM_SORT_ASCENDING : FM_SORT_DESCENDING;
+        fm_folder_model_set_sort(model, FM_FOLDER_MODEL_COL_DEFAULT, mode);
+    }
 }
 
 static void on_ui_destroy(gpointer ui_ptr)
@@ -1073,23 +1197,27 @@ GtkMenu* fm_folder_view_add_popup(FmFolderView* fv, GtkWindow* parent,
                                   FmFolderViewUpdatePopup update_popup)
 {
     FmFolderViewInterface* iface;
+    FmFolderModel* model;
     GtkUIManager* ui;
     GtkActionGroup* act_grp;
     GtkMenu* popup;
     GtkAction* act;
     GtkAccelGroup* accel_grp;
     gboolean show_hidden;
-    GtkSortType type;
-    FmFolderModelViewCol by;
+    FmSortMode mode;
+    GtkSortType type = (GtkSortType)-1;
+    FmFolderModelCol by = (FmFolderModelCol)-1;
 
     iface = FM_FOLDER_VIEW_GET_IFACE(fv);
     show_hidden = iface->get_show_hidden(fv);
-    iface->get_sort(fv, &type, &by);
+    model = iface->get_model(fv);
+    if(fm_folder_model_get_sort(model, &by, &mode))
+        type = FM_SORT_IS_ASCENDING(mode) ? GTK_SORT_ASCENDING : GTK_SORT_DESCENDING;
 
     /* init popup from XML string */
     ui = gtk_ui_manager_new();
-    act_grp = gtk_action_group_new("Desktop");
-    gtk_action_group_set_translation_domain(act_grp, NULL);
+    act_grp = gtk_action_group_new("Folder");
+    gtk_action_group_set_translation_domain(act_grp, GETTEXT_PACKAGE);
     gtk_action_group_add_actions(act_grp, folder_popup_actions,
                                  G_N_ELEMENTS(folder_popup_actions), fv);
     gtk_action_group_add_toggle_actions(act_grp, folder_toggle_actions,
@@ -1136,18 +1264,17 @@ GtkMenu* fm_folder_view_add_popup(FmFolderView* fv, GtkWindow* parent,
  * before this call.
  *
  * Implemented actions are:
- * <itemizedlist>
- * <listitem><para>Cut      : cut files into clipboard</para></listitem>
- * <listitem><para>Copy     : copy files into clipboard</para></listitem>
- * <listitem><para>Paste    : paste files from clipboard</para></listitem>
- * <listitem><para>Del      : move files into trash bin</para></listitem>
- * <listitem><para>Remove   : delete files from filesystem</para></listitem>
- * <listitem><para>SelAll   : select all</para></listitem>
- * <listitem><para>InvSel   : invert selection</para></listitem>
- * <listitem><para>Rename   : rename the folder</para></listitem>
- * <listitem><para>Prop     : folder properties dialog</para></listitem>
- * <listitem><para>FileProp : file properties dialog</para></listitem>
- * </itemizedlist>
+ * - Cut       : cut files into clipboard
+ * - Copy      : copy files into clipboard
+ * - Paste     : paste files from clipboard
+ * - Del       : move files into trash bin
+ * - Remove    : delete files from filesystem
+ * - SelAll    : select all
+ * - InvSel    : invert selection
+ * - Rename    : rename the folder
+ * - Prop      : folder properties dialog
+ * - FileProp  : file properties dialog
+ * - NewFolder : create new folder here
  *
  * See also: fm_folder_view_add_popup().
  *
@@ -1224,23 +1351,6 @@ void fm_folder_view_set_active(FmFolderView* fv, gboolean set)
     }
 }
 
-/* FIXME: move this near fm_launch_paths() */
-static FmJobErrorAction on_query_target_info_error(FmJob* job, GError* err, FmJobErrorSeverity severity, GtkWindow* win)
-{
-    if(err->domain == G_IO_ERROR)
-    {
-        if(err->code == G_IO_ERROR_NOT_MOUNTED)
-        {
-            if(fm_mount_path(win, fm_file_info_job_get_current(FM_FILE_INFO_JOB(job)), TRUE))
-                return FM_JOB_RETRY;
-        }
-        else if(err->code == G_IO_ERROR_FAILED_HANDLED)
-            return FM_JOB_CONTINUE;
-    }
-    fm_show_error(win, NULL, err->message);
-    return FM_JOB_CONTINUE;
-}
-
 /**
  * fm_folder_view_item_clicked
  * @fv: the folder view widget
@@ -1250,9 +1360,10 @@ static FmJobErrorAction on_query_target_info_error(FmJob* job, GError* err, FmJo
  * Handles left click and right click in folder area. If some item was
  * left-clicked then fm_folder_view_item_clicked() tries to launch it.
  * If some item was right-clicked then opens file menu (applying the
- * @update_popup before opening if it's not %NULL). If it was right-click
- * on empty space of folder view (so @path is NULL) then opens folder
- * popup that was created by fm_folder_view_add_popup(). After that
+ * update_popup returned by get_custom_menu_callbacks interface function
+ * before opening it if it's not %NULL). If it was right-click on empty
+ * space of folder view (so @path is %NULL) then opens folder popup
+ * menu that was created by fm_folder_view_add_popup(). After that
  * emits the #FmFolderView::clicked signal.
  *
  * If open_folders callback from interface function get_custom_menu_callbacks
@@ -1284,7 +1395,7 @@ void fm_folder_view_item_clicked(FmFolderView* fv, GtkTreePath* path,
     {
         model = GTK_TREE_MODEL(iface->get_model(fv));
         if(gtk_tree_model_get_iter(model, &it, path))
-            gtk_tree_model_get(model, &it, COL_FILE_INFO, &fi, -1);
+            gtk_tree_model_get(model, &it, FM_FOLDER_MODEL_COL_INFO, &fi, -1);
     }
     else
         fi = NULL;
@@ -1302,25 +1413,8 @@ void fm_folder_view_item_clicked(FmFolderView* fv, GtkTreePath* path,
         if(target && !fm_file_info_is_symlink(fi))
         {
             /* symlinks also has fi->target, but we only handle shortcuts here. */
-            FmFileInfo* target_fi;
             FmPath* real_path = fm_path_new_for_str(target);
-            /* query the info of target */
-            /* FIXME: all this should be done in fm_launch_paths() instead so
-               everything should be simplified to fm_launch_path_simple(...) */
-            FmFileInfoJob* job = fm_file_info_job_new(NULL, 0);
-            fm_file_info_job_add(job, real_path);
-            g_signal_connect(job, "error", G_CALLBACK(on_query_target_info_error), win);
-            fm_job_run_sync_with_mainloop(FM_JOB(job));
-            g_signal_handlers_disconnect_by_func(job, on_query_target_info_error, win);
-            target_fi = fm_file_info_list_peek_head(job->file_infos);
-            if(target_fi)
-                fm_file_info_ref(target_fi);
-            g_object_unref(job);
-            if(target_fi)
-            {
-                fm_launch_path_simple(win, NULL, real_path, open_folders, win);
-                fm_file_info_unref(target_fi);
-            }
+            fm_launch_path_simple(win, NULL, real_path, open_folders, win);
             fm_path_unref(real_path);
         }
         else
@@ -1332,6 +1426,13 @@ void fm_folder_view_item_clicked(FmFolderView* fv, GtkTreePath* path,
             FmFileMenu* menu;
             FmFileInfoList* files = iface->dup_selected_files(fv);
 
+            /* workaround on ExoTreeView bug */
+            if(files == NULL)
+            {
+                on_menu(NULL, fv);
+                goto send_signal;
+            }
+
             menu = fm_file_menu_new_for_files(win, files, fm_folder_view_get_cwd(fv), TRUE);
             fm_file_menu_set_folder_func(menu, open_folders, win);
 
@@ -1401,3 +1502,42 @@ void fm_folder_view_chdir(FmFolderView* fv, FmPath* path)
     g_signal_emit(fv, signals[CHDIR], 0, path);
 }
 #endif
+
+/* FIXME: make VTable entries after ABI change! */
+gboolean _fm_standard_view_set_columns(FmFolderView* fv, const GSList* cols);
+GSList* _fm_standard_view_get_columns(FmFolderView* fv);
+
+/**
+ * fm_folder_view_set_columns
+ * @fv: the folder view widget to apply
+ * @cols: (element-type FmFolderViewColumnInfo): new list of column infos
+ *
+ * Changes composition (rendering) of folder view @fv in accordance to
+ * new list of column infos.
+ *
+ * Returns: %TRUE in case of success.
+ *
+ * Since: 1.0.2
+ */
+gboolean fm_folder_view_set_columns(FmFolderView* fv, const GSList* cols)
+{
+    // FIXME: call via VTable!
+    return _fm_standard_view_set_columns(fv, cols);
+}
+
+/**
+ * fm_folder_view_get_columns
+ * @fv: the folder view widget to query
+ *
+ * Retrieves current composition of @fv as list of column infos. Returned
+ * list should be freed with g_slist_free() after usage.
+ *
+ * Returns: (transfer container) (element-type FmFolderViewColumnInfo): list of column infos.
+ *
+ * Since: 1.0.2
+ */
+GSList* fm_folder_view_get_columns(FmFolderView* fv)
+{
+    // FIXME: call via VTable!
+    return _fm_standard_view_get_columns(fv);
+}
