@@ -28,8 +28,7 @@
 #include <gio/gio.h>
 #include <stdarg.h>
 
-#include "fm-path.h"
-#include "fm-utils.h"
+#include "fm-file-info.h"
 #include "fm-file-launcher.h"
 
 G_BEGIN_DECLS
@@ -42,6 +41,8 @@ gboolean fm_launch_path_simple(GtkWindow* parent, GAppLaunchContext* ctx, FmPath
 
 gboolean fm_launch_desktop_entry_simple(GtkWindow* parent, GAppLaunchContext* ctx, FmFileInfo* entry, FmPathList* files);
 
+gboolean fm_launch_search_simple(GtkWindow* parent, GAppLaunchContext* ctx, const GList* paths, FmLaunchFolderFunc func, gpointer user_data);
+
 G_END_DECLS
 
 #endif /* __FM_GTK_UTILS_H__ */
