@@ -38,11 +38,8 @@
 #include <glib/gi18n-lib.h>
 #include "gtk-compat.h"
 
-#include "fm-marshal.h"
-#include "fm-config.h"
+#include "fm.h"
 #include "fm-standard-view.h"
-#include "fm-folder.h"
-#include "fm-folder-model.h"
 #include "fm-gtk-marshal.h"
 #include "fm-cell-renderer-text.h"
 #include "fm-cell-renderer-pixbuf.h"
@@ -51,6 +48,8 @@
 #include "exo/exo-icon-view.h"
 #include "exo/exo-tree-view.h"
 
+#include "fm-dnd-src.h"
+#include "fm-dnd-dest.h"
 #include "fm-dnd-auto-scroll.h"
 
 struct _FmStandardView
@@ -59,14 +58,13 @@ struct _FmStandardView
 
     FmStandardViewMode mode;
     GtkSelectionMode sel_mode;
-    GtkSortType sort_type;
-    FmFolderModelViewCol sort_by;
 
     gboolean show_hidden;
 
     GtkWidget* view; /* either ExoIconView or ExoTreeView */
     FmFolderModel* model; /* FmStandardView doesn't use abstract GtkTreeModel! */
-    FmCellRendererPixbuf* renderer_pixbuf; /* reference is bound to GtkWidget */
+    FmCellRendererPixbuf* renderer_pixbuf;
+    FmCellRendererText* renderer_text;
     guint icon_size_changed_handler;
 
     FmDndSrc* dnd_src; /* dnd source manager */
@@ -126,6 +124,18 @@ static void cancel_pending_row_activated(FmStandardView* fv);
 //static void on_folder_unmounted(FmFolder* folder, FmFolderView* fv);
 //static void on_folder_removed(FmFolder* folder, FmFolderView* fv);
 
+static FmFolderViewColumnInfo* _sv_column_info_new(FmFolderModelCol col_id)
+{
+    FmFolderViewColumnInfo* info = g_slice_new0(FmFolderViewColumnInfo);
+    info->col_id = col_id;
+    return info;
+}
+
+static void _sv_column_info_free(gpointer info)
+{
+    g_slice_free(FmFolderViewColumnInfo, info);
+}
+
 static void fm_standard_view_class_init(FmStandardViewClass *klass)
 {
     GObjectClass *g_object_class;
@@ -164,6 +174,9 @@ static gboolean on_idle_tree_view_row_activated(gpointer user_data)
 {
     FmStandardView* fv = (FmStandardView*)user_data;
     GtkTreePath* path;
+    GDK_THREADS_ENTER();
+    if(g_source_is_destroyed(g_main_current_source()))
+        goto _end;
     if(gtk_tree_row_reference_valid(fv->activated_row_ref))
     {
         path = gtk_tree_row_reference_get_path(fv->activated_row_ref);
@@ -173,6 +186,8 @@ static gboolean on_idle_tree_view_row_activated(gpointer user_data)
     gtk_tree_row_reference_free(fv->activated_row_ref);
     fv->activated_row_ref = NULL;
     fv->row_activated_idle = 0;
+_end:
+    GDK_THREADS_LEAVE();
     return FALSE;
 }
 
@@ -192,7 +207,7 @@ static void on_tree_view_row_activated(GtkTreeView* tv, GtkTreePath* path, GtkTr
     if(fv->model)
     {
         fv->activated_row_ref = gtk_tree_row_reference_new(GTK_TREE_MODEL(fv->model), path);
-        g_idle_add(on_idle_tree_view_row_activated, fv);
+        fv->row_activated_idle = g_idle_add(on_idle_tree_view_row_activated, fv);
     }
 }
 
@@ -209,11 +224,9 @@ static void fm_standard_view_init(FmStandardView *self)
     self->dnd_src = fm_dnd_src_new(NULL);
     g_signal_connect(self->dnd_src, "data-get", G_CALLBACK(on_dnd_src_data_get), self);
 
-    self->dnd_dest = fm_dnd_dest_new(NULL);
+    self->dnd_dest = fm_dnd_dest_new_with_handlers(NULL);
 
     self->mode = -1;
-    self->sort_type = GTK_SORT_ASCENDING;
-    self->sort_by = COL_FILE_NAME;
 }
 
 /**
@@ -243,10 +256,11 @@ FmFolderView* fm_folder_view_new(guint mode)
  * Since: 1.0.1
  */
 FmStandardView* fm_standard_view_new(FmStandardViewMode mode,
-                                        FmFolderViewUpdatePopup update_popup,
-                                        FmLaunchFolderFunc open_folders)
+                                     FmFolderViewUpdatePopup update_popup,
+                                     FmLaunchFolderFunc open_folders)
 {
     FmStandardView* fv = (FmStandardView*)g_object_new(FM_STANDARD_VIEW_TYPE, NULL);
+
     fm_standard_view_set_mode(fv, mode);
     fv->update_popup = update_popup;
     fv->open_folders = open_folders;
@@ -278,6 +292,18 @@ static void fm_standard_view_dispose(GObject *object)
     if(G_LIKELY(self->view))
         unset_view(self);
 
+    if(self->renderer_pixbuf)
+    {
+        g_object_unref(self->renderer_pixbuf);
+        self->renderer_pixbuf = NULL;
+    }
+
+    if(self->renderer_text)
+    {
+        g_object_unref(self->renderer_text);
+        self->renderer_text = NULL;
+    }
+
     if(self->cached_selected_files)
     {
         fm_file_info_list_unref(self->cached_selected_files);
@@ -322,6 +348,7 @@ static void fm_standard_view_dispose(GObject *object)
 static void set_icon_size(FmStandardView* fv, guint icon_size)
 {
     FmCellRendererPixbuf* render = fv->renderer_pixbuf;
+
     fm_cell_renderer_pixbuf_set_fixed_size(render, icon_size, icon_size);
 
     if(!fv->model)
@@ -331,12 +358,17 @@ static void set_icon_size(FmStandardView* fv, guint icon_size)
 
     if( fv->mode != FM_FV_LIST_VIEW ) /* this is an ExoIconView */
     {
-        /* FIXME: reset ExoIconView item sizes */
+        /* set row spacing in range 2...12 pixels */
+        gint c_size = MIN(12, 2 + icon_size / 8);
+        exo_icon_view_set_row_spacing(EXO_ICON_VIEW(fv->view), c_size);
     }
 }
 
 static void on_big_icon_size_changed(FmConfig* cfg, FmStandardView* fv)
 {
+    guint item_width = cfg->big_icon_size + 40;
+    /* reset ExoIconView item text sizes */
+    g_object_set((GObject*)fv->renderer_text, "wrap-width", item_width, NULL);
     set_icon_size(fv, cfg->big_icon_size);
 }
 
@@ -347,37 +379,14 @@ static void on_small_icon_size_changed(FmConfig* cfg, FmStandardView* fv)
 
 static void on_thumbnail_size_changed(FmConfig* cfg, FmStandardView* fv)
 {
+    guint item_width = MAX(cfg->thumbnail_size, 96);
+    /* reset ExoIconView item text sizes */
+    g_object_set((GObject*)fv->renderer_text, "wrap-width", item_width, NULL);
     /* FIXME: thumbnail and icons should have different sizes */
     /* maybe a separate API: fm_folder_model_set_thumbnail_size() */
     set_icon_size(fv, cfg->thumbnail_size);
 }
 
-static void on_drag_data_received(GtkWidget *dest_widget,
-                                    GdkDragContext *drag_context,
-                                    gint x,
-                                    gint y,
-                                    GtkSelectionData *sel_data,
-                                    guint info,
-                                    guint time,
-                                    FmStandardView* fv )
-{
-    fm_dnd_dest_drag_data_received(fv->dnd_dest, drag_context, x, y, sel_data, info, time);
-}
-
-static gboolean on_drag_drop(GtkWidget *dest_widget,
-                               GdkDragContext *drag_context,
-                               gint x,
-                               gint y,
-                               guint time,
-                               FmStandardView* fv)
-{
-    gboolean ret = FALSE;
-    GdkAtom target = gtk_drag_dest_find_target(dest_widget, drag_context, NULL);
-    if(target != GDK_NONE)
-        ret = fm_dnd_dest_drag_drop(fv->dnd_dest, drag_context, target, x, y, time);
-    return ret;
-}
-
 static GtkTreePath* get_drop_path_list_view(FmStandardView* fv, gint x, gint y)
 {
     GtkTreePath* tp = NULL;
@@ -386,7 +395,7 @@ static GtkTreePath* get_drop_path_list_view(FmStandardView* fv, gint x, gint y)
     gtk_tree_view_convert_widget_to_bin_window_coords(GTK_TREE_VIEW(fv->view), x, y, &x, &y);
     if(gtk_tree_view_get_path_at_pos(GTK_TREE_VIEW(fv->view), x, y, &tp, &col, NULL, NULL))
     {
-        if(gtk_tree_view_column_get_sort_column_id(col)!=COL_FILE_NAME)
+        if(gtk_tree_view_column_get_sort_column_id(col)!=FM_FOLDER_MODEL_COL_NAME)
         {
             gtk_tree_path_free(tp);
             tp = NULL;
@@ -432,7 +441,7 @@ static gboolean on_drag_motion(GtkWidget *dest_widget,
             if(gtk_tree_model_get_iter(GTK_TREE_MODEL(fv->model), &it, tp))
             {
                 FmFileInfo* fi;
-                gtk_tree_model_get(GTK_TREE_MODEL(fv->model), &it, COL_FILE_INFO, &fi, -1);
+                gtk_tree_model_get(GTK_TREE_MODEL(fv->model), &it, FM_FOLDER_MODEL_COL_INFO, &fi, -1);
                 fm_dnd_dest_set_dest_file(fv->dnd_dest, fi);
             }
             gtk_tree_path_free(tp);
@@ -456,30 +465,24 @@ static gboolean on_drag_motion(GtkWidget *dest_widget,
     return ret;
 }
 
-static void on_drag_leave(GtkWidget *dest_widget,
-                                GdkDragContext *drag_context,
-                                guint time,
-                                FmStandardView* fv)
-{
-    fm_dnd_dest_drag_leave(fv->dnd_dest, drag_context, time);
-}
-
 static inline void create_icon_view(FmStandardView* fv, GList* sels)
 {
     GList *l;
     GtkCellRenderer* render;
     FmFolderModel* model = fv->model;
-    int icon_size = 0;
+    int icon_size = 0, item_width;
 
     fv->view = exo_icon_view_new();
 
-    fv->renderer_pixbuf = fm_cell_renderer_pixbuf_new();
+    if(fv->renderer_pixbuf)
+        g_object_unref(fv->renderer_pixbuf);
+    fv->renderer_pixbuf = g_object_ref_sink(fm_cell_renderer_pixbuf_new());
     render = (GtkCellRenderer*)fv->renderer_pixbuf;
 
     g_object_set((GObject*)render, "follow-state", TRUE, NULL );
     gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(fv->view), render, TRUE);
-    gtk_cell_layout_add_attribute(GTK_CELL_LAYOUT(fv->view), render, "pixbuf", COL_FILE_ICON );
-    gtk_cell_layout_add_attribute(GTK_CELL_LAYOUT(fv->view), render, "info", COL_FILE_INFO );
+    gtk_cell_layout_add_attribute(GTK_CELL_LAYOUT(fv->view), render, "pixbuf", FM_FOLDER_MODEL_COL_ICON );
+    gtk_cell_layout_add_attribute(GTK_CELL_LAYOUT(fv->view), render, "info", FM_FOLDER_MODEL_COL_INFO );
 
     if(fv->mode == FM_FV_COMPACT_VIEW) /* compact view */
     {
@@ -508,16 +511,16 @@ static inline void create_icon_view(FmStandardView* fv, GList* sels)
                 fm_folder_model_set_icon_size(model, icon_size);
 
             render = fm_cell_renderer_text_new();
-            /* FIXME: set the sizes of cells according to iconsize */
+            item_width = icon_size + 40;
             g_object_set((GObject*)render,
                          "wrap-mode", PANGO_WRAP_WORD_CHAR,
-                         "wrap-width", 90,
+                         "wrap-width", item_width,
+                         "max-height", 70,
                          "alignment", PANGO_ALIGN_CENTER,
                          "xalign", 0.5,
                          "yalign", 0.0,
                          NULL );
             exo_icon_view_set_column_spacing( (ExoIconView*)fv->view, 4 );
-            exo_icon_view_set_item_width ( (ExoIconView*)fv->view, 110 );
         }
         else
         {
@@ -528,23 +531,25 @@ static inline void create_icon_view(FmStandardView* fv, GList* sels)
                 fm_folder_model_set_icon_size(model, icon_size);
 
             render = fm_cell_renderer_text_new();
-            /* FIXME: set the sizes of cells according to iconsize */
+            item_width = MAX(icon_size, 96);
             g_object_set((GObject*)render,
                          "wrap-mode", PANGO_WRAP_WORD_CHAR,
-                         "wrap-width", 180,
+                         "wrap-width", item_width,
+                         "max-height", 90,
                          "alignment", PANGO_ALIGN_CENTER,
                          "xalign", 0.5,
                          "yalign", 0.0,
                          NULL );
             exo_icon_view_set_column_spacing( (ExoIconView*)fv->view, 8 );
-            exo_icon_view_set_item_width ( (ExoIconView*)fv->view, 200 );
         }
     }
     gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(fv->view), render, TRUE);
     gtk_cell_layout_add_attribute(GTK_CELL_LAYOUT(fv->view), render,
-                                "text", COL_FILE_NAME );
-    exo_icon_view_set_item_width((ExoIconView*)fv->view, 96);
-    exo_icon_view_set_search_column((ExoIconView*)fv->view, COL_FILE_NAME);
+                                "text", FM_FOLDER_MODEL_COL_NAME );
+    if(fv->renderer_text)
+        g_object_unref(fv->renderer_text);
+    fv->renderer_text = g_object_ref_sink(render);
+    exo_icon_view_set_search_column((ExoIconView*)fv->view, FM_FOLDER_MODEL_COL_NAME);
     g_signal_connect(fv->view, "item-activated", G_CALLBACK(on_icon_view_item_activated), fv);
     g_signal_connect(fv->view, "selection-changed", G_CALLBACK(on_sel_changed), fv);
     exo_icon_view_set_model((ExoIconView*)fv->view, (GtkTreeModel*)fv->model);
@@ -556,65 +561,134 @@ static inline void create_icon_view(FmStandardView* fv, GList* sels)
         exo_icon_view_select_path((ExoIconView*)fv->view, l->data);
 }
 
-static inline void create_list_view(FmStandardView* fv, GList* sels)
+static void on_column_width_changed(GtkTreeViewColumn* col, FmStandardView* view)
+{
+    /* FmFolderViewColumnInfo* info = g_object_get_qdata(G_OBJECT(col), fm_qdata_id);
+    g_debug("column width changed: id %u mode %u: %d", info->col_id,
+            gtk_tree_view_column_get_sizing(col), gtk_tree_view_column_get_width(col)); */
+    return;
+#if 0
+    int width;
+    GList* cols = gtk_tree_view_get_columns(GTK_TREE_VIEW(view->view));
+    int pos = g_list_index(cols, col);
+    g_list_free(cols);
+    width = gtk_tree_view_column_get_width(col);
+
+    /* only handle it if the width really changed */
+    if(width != view->column_widths[pos])
+    {
+        // FIXME: what if column_widths = NULL?
+        view->column_widths[pos] = width;
+        // g_signal_emit(view, signals[COLUMN_WIDTH_CHANGED], 0);
+    }
+#endif
+}
+
+static GtkTreeViewColumn* create_list_view_column(FmStandardView* fv,
+                                                  FmFolderViewColumnInfo *set)
 {
     GtkTreeViewColumn* col;
+    GtkCellRenderer* render;
+    const char* title;
+    FmFolderViewColumnInfo* info;
+    FmFolderModelCol col_id;
+
+    g_return_val_if_fail(set != NULL, NULL); /* invalid arg */
+    col_id = set->col_id;
+    title = fm_folder_model_col_get_title(fv->model, col_id);
+    g_return_val_if_fail(title != NULL, NULL); /* invalid column */
+
+    /* g_debug("adding column id=%u", col_id); */
+    col = gtk_tree_view_column_new();
+    render = gtk_cell_renderer_text_new();
+    gtk_tree_view_column_set_title(col, title);
+    info = _sv_column_info_new(col_id);
+    /* TODO: update other data from set - width for example */
+    g_object_set_qdata_full(G_OBJECT(col), fm_qdata_id, info, _sv_column_info_free);
+
+    if(G_UNLIKELY(col_id == FM_FOLDER_MODEL_COL_NAME)) /* special handling for Name column */
+    {
+        gtk_tree_view_column_pack_start(col, GTK_CELL_RENDERER(fv->renderer_pixbuf), FALSE);
+        gtk_tree_view_column_set_attributes(col, GTK_CELL_RENDERER(fv->renderer_pixbuf),
+                                            "pixbuf", FM_FOLDER_MODEL_COL_ICON,
+                                            "info", FM_FOLDER_MODEL_COL_INFO, NULL);
+        g_object_set(render, "ellipsize", PANGO_ELLIPSIZE_END, NULL);
+        gtk_tree_view_column_set_expand(col, TRUE);
+        gtk_tree_view_column_set_sizing(col, GTK_TREE_VIEW_COLUMN_FIXED);
+        gtk_tree_view_column_set_fixed_width(col, 200);
+    }
+    else
+    {
+        if(G_UNLIKELY(col_id == FM_FOLDER_MODEL_COL_SIZE))
+            g_object_set(render, "xalign", 1.0, NULL);
+    }
+
+    gtk_tree_view_column_pack_start(col, render, TRUE);
+    gtk_tree_view_column_set_attributes(col, render, "text", col_id, NULL);
+    gtk_tree_view_column_set_resizable(col, TRUE);
+    if(fm_folder_model_col_is_sortable(fv->model, col_id))
+        gtk_tree_view_column_set_sort_column_id(col, col_id);
+    gtk_tree_view_append_column(GTK_TREE_VIEW(fv->view), col);
+    if(G_UNLIKELY(col_id == FM_FOLDER_MODEL_COL_NAME))
+        /* only this column is activable */
+        exo_tree_view_set_activable_column((ExoTreeView*)fv->view, col);
+
+    g_signal_connect(col, "notify::width", G_CALLBACK(on_column_width_changed), fv);
+
+    return col;
+}
+
+static void _check_tree_columns_defaults(FmStandardView* fv)
+{
+    const FmFolderViewColumnInfo cols[] = {
+        {FM_FOLDER_MODEL_COL_NAME},
+        {FM_FOLDER_MODEL_COL_DESC},
+        {FM_FOLDER_MODEL_COL_SIZE},
+        {FM_FOLDER_MODEL_COL_MTIME} };
+    GSList* cols_list = NULL;
+    GList* tree_columns;
+    guint i;
+
+    tree_columns = gtk_tree_view_get_columns(GTK_TREE_VIEW(fv->view));
+    if(tree_columns != NULL) /* already set */
+    {
+        g_list_free(tree_columns);
+        return;
+    }
+    /* Set default columns to show in detailed list mode.
+     * FIXME: cols should be passed to fm_standard_view_new() as a parameter instead.
+     * This breaks API/ABI though. Let's do it later. */
+    for(i = 0; i < G_N_ELEMENTS(cols); i++)
+        cols_list = g_slist_append(cols_list, (gpointer)&cols[i]);
+    fm_folder_view_set_columns(FM_FOLDER_VIEW(fv), cols_list);
+    g_slist_free(cols_list);
+}
+
+static inline void create_list_view(FmStandardView* fv, GList* sels)
+{
     GtkTreeSelection* ts;
     GList *l;
-    GtkCellRenderer* render;
     FmFolderModel* model = fv->model;
     int icon_size = 0;
+
     fv->view = exo_tree_view_new();
 
-    fv->renderer_pixbuf = fm_cell_renderer_pixbuf_new();
-    render = (GtkCellRenderer*)fv->renderer_pixbuf;
+    if(fv->renderer_pixbuf)
+        g_object_unref(fv->renderer_pixbuf);
+    fv->renderer_pixbuf = g_object_ref_sink(fm_cell_renderer_pixbuf_new());
     fv->icon_size_changed_handler = g_signal_connect(fm_config, "changed::small_icon_size", G_CALLBACK(on_small_icon_size_changed), fv);
     icon_size = fm_config->small_icon_size;
     fm_cell_renderer_pixbuf_set_fixed_size(fv->renderer_pixbuf, icon_size, icon_size);
     if(model)
+    {
         fm_folder_model_set_icon_size(model, icon_size);
+        _check_tree_columns_defaults(fv);
+        gtk_tree_view_set_search_column(GTK_TREE_VIEW(fv->view),
+                                        FM_FOLDER_MODEL_COL_NAME);
+    }
 
+    gtk_tree_view_set_reorderable(GTK_TREE_VIEW(fv->view), TRUE);
     gtk_tree_view_set_rules_hint(GTK_TREE_VIEW(fv->view), TRUE);
-    col = gtk_tree_view_column_new();
-    gtk_tree_view_column_set_title(col, _("Name"));
-    gtk_tree_view_column_pack_start(col, render, FALSE);
-    gtk_tree_view_column_set_attributes(col, render,
-                                        "pixbuf", COL_FILE_ICON,
-                                        "info", COL_FILE_INFO, NULL);
-    render = gtk_cell_renderer_text_new();
-    g_object_set(render, "ellipsize", PANGO_ELLIPSIZE_END, NULL);
-    gtk_tree_view_column_pack_start(col, render, TRUE);
-    gtk_tree_view_column_set_attributes(col, render, "text", COL_FILE_NAME, NULL);
-    gtk_tree_view_column_set_sort_column_id(col, COL_FILE_NAME);
-    gtk_tree_view_column_set_expand(col, TRUE);
-    gtk_tree_view_column_set_resizable(col, TRUE);
-    gtk_tree_view_column_set_sizing(col, GTK_TREE_VIEW_COLUMN_FIXED);
-    gtk_tree_view_column_set_fixed_width(col, 200);
-    gtk_tree_view_append_column(GTK_TREE_VIEW(fv->view), col);
-    /* only this column is activable */
-    exo_tree_view_set_activable_column((ExoTreeView*)fv->view, col);
-
-    render = gtk_cell_renderer_text_new();
-    col = gtk_tree_view_column_new_with_attributes(_("Description"), render, "text", COL_FILE_DESC, NULL);
-    gtk_tree_view_column_set_resizable(col, TRUE);
-    gtk_tree_view_column_set_sort_column_id(col, COL_FILE_DESC);
-    gtk_tree_view_append_column(GTK_TREE_VIEW(fv->view), col);
-
-    render = gtk_cell_renderer_text_new();
-    g_object_set(render, "xalign", 1.0, NULL);
-    col = gtk_tree_view_column_new_with_attributes(_("Size"), render, "text", COL_FILE_SIZE, NULL);
-    gtk_tree_view_column_set_sort_column_id(col, COL_FILE_SIZE);
-    gtk_tree_view_column_set_resizable(col, TRUE);
-    gtk_tree_view_append_column(GTK_TREE_VIEW(fv->view), col);
-
-    render = gtk_cell_renderer_text_new();
-    col = gtk_tree_view_column_new_with_attributes(_("Modified"), render, "text", COL_FILE_MTIME, NULL);
-    gtk_tree_view_column_set_resizable(col, TRUE);
-    gtk_tree_view_column_set_sort_column_id(col, COL_FILE_MTIME);
-    gtk_tree_view_append_column(GTK_TREE_VIEW(fv->view), col);
-
-    gtk_tree_view_set_search_column(GTK_TREE_VIEW(fv->view), COL_FILE_NAME);
-
     gtk_tree_view_set_rubber_banding(GTK_TREE_VIEW(fv->view), TRUE);
     exo_tree_view_set_single_click((ExoTreeView*)fv->view, fm_config->single_click);
     exo_tree_view_set_single_click_timeout((ExoTreeView*)fv->view, SINGLE_CLICK_TIMEOUT);
@@ -623,7 +697,7 @@ static inline void create_list_view(FmStandardView* fv, GList* sels)
     g_signal_connect(fv->view, "row-activated", G_CALLBACK(on_tree_view_row_activated), fv);
     g_signal_connect(ts, "changed", G_CALLBACK(on_sel_changed), fv);
     /*cancel_pending_row_activated(fv);*/ /* FIXME: is this needed? */
-    gtk_tree_view_set_model(GTK_TREE_VIEW(fv->view), GTK_TREE_MODEL(fv->model));
+    gtk_tree_view_set_model(GTK_TREE_VIEW(fv->view), GTK_TREE_MODEL(model));
     gtk_tree_selection_set_mode(ts, fv->sel_mode);
     for(l = sels;l;l=l->next)
         gtk_tree_selection_select_path(ts, (GtkTreePath*)l->data);
@@ -642,9 +716,6 @@ static void unset_view(FmStandardView* fv)
         g_signal_handlers_disconnect_by_func(fv->view, on_sel_changed, fv);
     /* these signals connected by fm_standard_view_set_mode() */
     g_signal_handlers_disconnect_by_func(fv->view, on_drag_motion, fv);
-    g_signal_handlers_disconnect_by_func(fv->view, on_drag_leave, fv);
-    g_signal_handlers_disconnect_by_func(fv->view, on_drag_drop, fv);
-    g_signal_handlers_disconnect_by_func(fv->view, on_drag_data_received, fv);
     g_signal_handlers_disconnect_by_func(fv->view, on_btn_pressed, fv);
 
     fm_dnd_unset_dest_auto_scroll(fv->view);
@@ -804,9 +875,6 @@ void fm_standard_view_set_mode(FmStandardView* fv, FmStandardViewMode mode)
         fm_dnd_src_set_widget(fv->dnd_src, fv->view);
         fm_dnd_dest_set_widget(fv->dnd_dest, fv->view);
         g_signal_connect_after(fv->view, "drag-motion", G_CALLBACK(on_drag_motion), fv);
-        g_signal_connect(fv->view, "drag-leave", G_CALLBACK(on_drag_leave), fv);
-        g_signal_connect(fv->view, "drag-drop", G_CALLBACK(on_drag_drop), fv);
-        g_signal_connect(fv->view, "drag-data-received", G_CALLBACK(on_drag_data_received), fv);
         /* connecting it after sometimes conflicts with system configuration
            (bug #3559831) so we just hope here it will be handled in order
            of connecting, i.e. after ExoIconView or ExoTreeView handler */
@@ -888,20 +956,6 @@ static GtkSelectionMode fm_standard_view_get_selection_mode(FmFolderView* ffv)
     return fv->sel_mode;
 }
 
-static void fm_standard_view_set_sort(FmFolderView* ffv, GtkSortType type, FmFolderModelViewCol by)
-{
-    FmStandardView* fv = FM_STANDARD_VIEW(ffv);
-    fv->sort_type = type;
-    fv->sort_by = by;
-}
-
-static void fm_standard_view_get_sort(FmFolderView* ffv, GtkSortType* type, FmFolderModelViewCol* by)
-{
-    FmStandardView* fv = FM_STANDARD_VIEW(ffv);
-    *type = fv->sort_type;
-    *by = fv->sort_by;
-}
-
 static void fm_standard_view_set_show_hidden(FmFolderView* ffv, gboolean show)
 {
     FmStandardView* fv = FM_STANDARD_VIEW(ffv);
@@ -936,15 +990,6 @@ static GList* fm_standard_view_get_selected_tree_paths(FmStandardView* fv)
     return sels;
 }
 
-/**
- * fm_folder_view_dup_selected_files
- * @fv: a FmFolderView object
- *
- * Return value: An referenced FmFileInfoList containing FmFileInfos of
- * the currently selected files. The list should be freed after usage with
- * fm_file_info_list_unref(). If there are no files selected then return
- * value is %NULL.
- **/
 static inline FmFileInfoList* fm_standard_view_get_selected_files(FmStandardView* fv)
 {
     /* don't generate the data again if we have it cached. */
@@ -961,7 +1006,7 @@ static inline FmFileInfoList* fm_standard_view_get_selected_files(FmStandardView
                 GtkTreeIter it;
                 GtkTreePath* tp = (GtkTreePath*)l->data;
                 gtk_tree_model_get_iter(GTK_TREE_MODEL(fv->model), &it, l->data);
-                gtk_tree_model_get(GTK_TREE_MODEL(fv->model), &it, COL_FILE_INFO, &fi, -1);
+                gtk_tree_model_get(GTK_TREE_MODEL(fv->model), &it, FM_FOLDER_MODEL_COL_INFO, &fi, -1);
                 gtk_tree_path_free(tp);
                 next = l->next;
                 l->data = fm_file_info_ref( fi );
@@ -979,15 +1024,6 @@ static FmFileInfoList* fm_standard_view_dup_selected_files(FmFolderView* ffv)
     return fm_file_info_list_ref(fm_standard_view_get_selected_files(fv));
 }
 
-/**
- * fm_folder_view_dup_selected_file_paths
- * @fv: a FmFolderView object
- *
- * Return value: An referenced FmPathList containing FmPaths of the
- * the currently selected files. The list should be freed after usage with
- * fm_path_list_unref(). If there are no files selected then return value
- * is %NULL.
- **/
 static FmPathList* fm_standard_view_dup_selected_file_paths(FmFolderView* ffv)
 {
     FmStandardView* fv = FM_STANDARD_VIEW(ffv);
@@ -1039,8 +1075,11 @@ static gboolean on_btn_pressed(GtkWidget* view, GdkEventButton* evt, FmStandardV
         /* special handling for ExoIconView */
         if(evt->button != 1)
         {
-            if(fv->mode==FM_FV_ICON_VIEW || fv->mode==FM_FV_COMPACT_VIEW || fv->mode==FM_FV_THUMBNAIL_VIEW)
+            switch(fv->mode)
             {
+            case FM_FV_ICON_VIEW:
+            case FM_FV_COMPACT_VIEW:
+            case FM_FV_THUMBNAIL_VIEW:
                 /* select the item on right click for ExoIconView */
                 if(exo_icon_view_get_item_at_pos(EXO_ICON_VIEW(view), evt->x, evt->y, &tp, NULL))
                 {
@@ -1058,10 +1097,10 @@ static gboolean on_btn_pressed(GtkWidget* view, GdkEventButton* evt, FmStandardV
                         exo_icon_view_set_cursor(EXO_ICON_VIEW(view), tp, NULL, FALSE);
                     }
                 }
-            }
-            else if( fv->mode == FM_FV_LIST_VIEW
-                     && evt->window == gtk_tree_view_get_bin_window(GTK_TREE_VIEW(view)))
-            {
+                break;
+            case FM_FV_LIST_VIEW:
+              if(evt->window == gtk_tree_view_get_bin_window(GTK_TREE_VIEW(view)))
+              {
                 /* special handling for ExoTreeView */
                 /* Fix #2986834: MAJOR PROBLEM: Deletes Wrong File Frequently. */
                 GtkTreeViewColumn* col;
@@ -1078,6 +1117,7 @@ static gboolean on_btn_pressed(GtkWidget* view, GdkEventButton* evt, FmStandardV
                         }
                     }
                 }
+              }
             }
         }
 
@@ -1127,11 +1167,9 @@ static void fm_standard_view_unselect_all(FmFolderView* ffv)
 static void on_dnd_src_data_get(FmDndSrc* ds, FmStandardView* fv)
 {
     FmFileInfoList* files = fm_standard_view_dup_selected_files(FM_FOLDER_VIEW(fv));
+    fm_dnd_src_set_files(ds, files);
     if(files)
-    {
-        fm_dnd_src_set_files(ds, files);
         fm_file_info_list_unref(files);
-    }
 }
 
 static gboolean on_sel_changed_real(FmStandardView* fv)
@@ -1161,11 +1199,18 @@ static gboolean on_sel_changed_real(FmStandardView* fv)
 static gboolean on_sel_changed_idle(gpointer user_data)
 {
     FmStandardView* fv = (FmStandardView*)user_data;
+    gboolean ret = FALSE;
 
+    GDK_THREADS_ENTER();
+    /* check if fv is destroyed already */
+    if(g_source_is_destroyed(g_main_current_source()))
+        goto _end;
     if(fv->sel_changed_pending) /* fast changing detected! continue... */
-        return on_sel_changed_real(fv);
+        ret = on_sel_changed_real(fv);
     fv->sel_changed_idle = 0;
-    return FALSE;
+_end:
+    GDK_THREADS_LEAVE();
+    return ret;
 }
 
 static void on_sel_changed(GObject* obj, FmStandardView* fv)
@@ -1203,7 +1248,7 @@ static void fm_standard_view_select_file_path(FmFolderView* ffv, FmPath* path)
         FmFolderModel* model = fv->model;
         GtkTreeIter it;
         if(fv->select_path &&
-           fm_folder_model_find_iter_by_filename(model, &it, path->name))
+           fm_folder_model_find_iter_by_filename(model, &it, fm_path_get_basename(path)))
             fv->select_path(model, fv->view, &it);
     }
 }
@@ -1249,6 +1294,7 @@ static void fm_standard_view_set_model(FmFolderView* ffv, FmFolderModel* model)
     {
     case FM_FV_LIST_VIEW:
         cancel_pending_row_activated(fv);
+        _check_tree_columns_defaults(fv);
         if(model)
         {
             icon_size = fm_config->small_icon_size;
@@ -1286,12 +1332,112 @@ static void fm_standard_view_set_model(FmFolderView* ffv, FmFolderModel* model)
         fv->model = NULL;
 }
 
+typedef struct
+{
+    GtkTreeViewColumn* col;
+    FmFolderViewColumnInfo* info;
+} _ColumnsCache;
+
+gboolean _fm_standard_view_set_columns(FmFolderView* fv, const GSList* cols)
+{
+    FmStandardView* view;
+    GtkTreeViewColumn *col, *last;
+    FmFolderViewColumnInfo* info;
+    _ColumnsCache* old_cols;
+    const GSList* l;
+    GList *cols_list, *ld;
+    guint i, n_cols;
+
+    if(!FM_IS_STANDARD_VIEW(fv))
+        return FALSE;
+    view = (FmStandardView*)fv;
+
+    if(view->mode != FM_FV_LIST_VIEW) /* other modes aren't supported now */
+        return FALSE;
+
+    cols_list = gtk_tree_view_get_columns(GTK_TREE_VIEW(view->view));
+    n_cols = g_list_length(cols_list);
+    if(n_cols > 0)
+    {
+        /* create more convenient for us list of columns */
+        old_cols = g_new(_ColumnsCache, n_cols);
+        for(ld = cols_list, i = 0; ld; ld = ld->next, i++)
+        {
+            col = ld->data; /* column */
+            info = g_object_get_qdata(G_OBJECT(col), fm_qdata_id); /* info */
+            old_cols[i].col = col;
+            old_cols[i].info = info;
+        }
+        g_list_free(cols_list);
+    }
+    last = NULL;
+    for(l = cols; l; l = l->next)
+    {
+        info = l->data;
+        /* find old one and move here */
+        for(i = 0; i < n_cols; i++)
+            if(old_cols[i].info && old_cols[i].info->col_id == info->col_id)
+                break;
+        if(i < n_cols)
+        {
+            /* we found it so just move it here */
+            col = old_cols[i].col;
+            old_cols[i].col = NULL; /* we removed it from its place */
+            old_cols[i].info = NULL; /* don't try to use it again */
+            /* TODO: update all other data - width for example */
+        }
+        else
+        {
+            /* if not found then append new one */
+            col = create_list_view_column(view, info);
+            if(col == NULL) /* failed! skipping it */
+                continue;
+        }
+        gtk_tree_view_move_column_after(GTK_TREE_VIEW(view->view), col, last);
+        last = col;
+    }
+
+    /* remove abandoned columns from view */
+    for(i = 0; i < n_cols; i++)
+        if(old_cols[i].col != NULL)
+            gtk_tree_view_remove_column(GTK_TREE_VIEW(view->view),
+                                        old_cols[i].col);
+    if(n_cols > 0)
+        g_free(old_cols);
+    return TRUE;
+}
+
+GSList* _fm_standard_view_get_columns(FmFolderView* fv)
+{
+    FmStandardView* view;
+    GSList* list;
+    GList *cols_list, *ld;
+
+    if(!FM_IS_STANDARD_VIEW(fv))
+        return NULL;
+    view = (FmStandardView*)fv;
+
+    if(view->mode != FM_FV_LIST_VIEW) /* other modes aren't supported now */
+        return NULL;
+
+    cols_list = gtk_tree_view_get_columns(GTK_TREE_VIEW(view->view));
+    if(cols_list == NULL)
+        return NULL;
+    list = NULL;
+    for(ld = cols_list; ld; ld = ld->next)
+    {
+        GtkTreeViewColumn *col = ld->data;
+        FmFolderViewColumnInfo* info = g_object_get_qdata(G_OBJECT(col), fm_qdata_id);
+        list = g_slist_append(list, info); /* info */
+    }
+    g_list_free(cols_list);
+    return list;
+}
+
 static void fm_standard_view_view_init(FmFolderViewInterface* iface)
 {
     iface->set_sel_mode = fm_standard_view_set_selection_mode;
     iface->get_sel_mode = fm_standard_view_get_selection_mode;
-    iface->set_sort = fm_standard_view_set_sort;
-    iface->get_sort = fm_standard_view_get_sort;
     iface->set_show_hidden = fm_standard_view_set_show_hidden;
     iface->get_show_hidden = fm_standard_view_get_show_hidden;
     iface->get_folder = fm_standard_view_get_folder;
@@ -1306,3 +1452,60 @@ static void fm_standard_view_view_init(FmFolderViewInterface* iface)
     iface->select_file_path = fm_standard_view_select_file_path;
     iface->get_custom_menu_callbacks = fm_standard_view_get_custom_menu_callbacks;
 }
+
+typedef struct
+{
+    const char* name;
+    FmStandardViewMode mode;
+} _ModeNames;
+
+static const _ModeNames view_mode_names[] =
+{
+    { "icon", FM_FV_ICON_VIEW },
+    { "compact", FM_FV_COMPACT_VIEW },
+    { "thumbnail", FM_FV_THUMBNAIL_VIEW },
+    { "list", FM_FV_LIST_VIEW }
+};
+
+/**
+ * fm_standard_view_mode_to_str
+ * @mode: mode id
+ *
+ * Retrieves string name of rendering @mode. That name may be used for
+ * config save or similar purposes. Returned data are owned by the
+ * implementation and should be not freed by caller.
+ *
+ * Returns: name associated with @mode.
+ *
+ * Since: 1.0.2
+ */
+const char* fm_standard_view_mode_to_str(FmStandardViewMode mode)
+{
+    guint i;
+
+    if(G_LIKELY(FM_STANDARD_VIEW_MODE_IS_VALID(mode)))
+        for(i = 0; i < G_N_ELEMENTS(view_mode_names); i++)
+            if(view_mode_names[i].mode == mode)
+                return view_mode_names[i].name;
+    return NULL;
+}
+
+/**
+ * fm_standard_view_mode_from_str
+ * @str: the name of mode
+ *
+ * Finds mode which have an associated name equal to @str.
+ *
+ * Returns: mode id or (FmStandardViewMode)-1 if no such mode exists.
+ *
+ * Since: 1.0.2
+ */
+FmStandardViewMode fm_standard_view_mode_from_str(const char* str)
+{
+    guint i;
+
+    for(i = 0; i < G_N_ELEMENTS(view_mode_names); i++)
+        if(strcmp(str, view_mode_names[i].name) == 0)
+            return view_mode_names[i].mode;
+    return (FmStandardViewMode)-1;
+}
