@@ -91,6 +91,9 @@ static gboolean _fm_file_ops_job_copy_file(FmFileOpsJob* job, GFile* src, GFileI
     guint32 mode;
     gboolean skip_dir_content = FALSE;
 
+    /* FIXME: g_file_get_child() failed? generate error! */
+    g_return_val_if_fail(dest != NULL, FALSE);
+
     if( G_LIKELY(inf) )
         g_object_ref(inf);
     else
@@ -658,7 +661,10 @@ gboolean _fm_file_ops_job_copy_run(FmFileOpsJob* job)
     dest_dir = fm_path_to_gfile(job->dest);
     /* get dummy file monitors for non-native filesystems */
     if( g_file_is_native(dest_dir) )
+    {
+        old_mon = NULL; /* to satisfy compiler */
         dest_mon = NULL;
+    }
     else
     {
         old_mon = job->dest_folder_mon;
@@ -672,7 +678,7 @@ gboolean _fm_file_ops_job_copy_run(FmFileOpsJob* job)
     {
         FmPath* path = FM_PATH(l->data);
         GFile* src = fm_path_to_gfile(path);
-        GFile* dest = g_file_get_child(dest_dir, path->name);
+        GFile* dest = g_file_get_child(dest_dir, fm_path_get_basename(path));
 
         if(!_fm_file_ops_job_copy_file(job, src, NULL, dest))
             ret = FALSE;
@@ -751,7 +757,10 @@ _retry_query_dest_info:
 
     /* get dummy file monitors for non-native filesystems */
     if( g_file_is_native(dest_dir) )
+    {
+        old_mon = NULL; /* to satisfy compiler */
         dest_mon = NULL;
+    }
     else
     {
         old_mon = job->dest_folder_mon;
@@ -766,7 +775,7 @@ _retry_query_dest_info:
     {
         FmPath* path = FM_PATH(l->data);
         GFile* src = fm_path_to_gfile(path);
-        GFile* dest = g_file_get_child(dest_dir, path->name);
+        GFile* dest = g_file_get_child(dest_dir, fm_path_get_basename(path));
 
         /* get dummy file monitors for non-native filesystems */
         job->src_folder_mon = NULL;
